import multiprocessing
import time

from bspider.agent import log
from bspider.core.api import BaseService, GetSuccess, PostSuccess, DeleteSuccess, Conflict, NotFound
from bspider.downloader.work import run_downloader
from bspider.parser.work import run_parser
from bspider.utils.system import System


class NodeService(BaseService):

    def __init__(self):
        """
        由于模块进程的特殊性，不能使用fork 方式产生多进程，必须采用spawn
        https://docs.python.org/3.6/library/multiprocessing.html?highlight=queue#multiprocessing.get_context
        """
        # 存储子worker对象
        self.module_list = {}
        self.mp_ctx = multiprocessing.get_context('spawn')

        self.system = System()

    def node_status(self):
        """报告 磁盘、处理器、内存使用状况"""
        return GetSuccess(
            msg='get node status success',
            data={
                'cpu_num': self.system.cpu_num,
                'cpu_percent': self.system.cpu_percent,
                'mem_size': self.system.mem_size,
                'mem_percent': self.system.mem_percent,
                'disk_size': self.system.disk_size,
                'disk_percent': self.system.disk_percent
            }
        )

    def start_worker(self, worker_id, worker_type, coroutine_num):
        unique_id = f'worker_{worker_id}'
        if unique_id in self.module_list:
            log.error(f'worker:worker_id->{worker_id} is already exist in this node')
            return Conflict(msg=f'worker:worker_id->{worker_id} is already exist in this node', errno=20001)

        if worker_type == 'downloader':
            func = run_downloader
        elif worker_type == 'parser':
            func = run_parser
        else:
            log.error(f'unknow worker type: {worker_type}')
            return Conflict(msg=f'unknow worker type: {worker_type}', errno=20002)

        worker = self.__start(func, unique_id, coroutine_num)
        if worker.is_alive():
            self.module_list[unique_id] = worker
            log.info(f'start work:{worker_id}-{worker_type} success')
            return PostSuccess(msg='worker process start success', data={'pid': worker.pid, 'status': 1})
        else:
            log.error(f'worker process start error. module start exec')
            return Conflict(msg=f'worker process start error', errno=20003)

    def __start(self, func, unique_id, coro_num):
        worker = self.mp_ctx.Process(name=unique_id, target=func, args=(unique_id, coro_num), daemon=True)
        worker.start()
        return worker

    def stop_worker(self, worker_id):
        unique_id = f'worker_{worker_id}'
        try:
            worker = self.module_list.pop(unique_id)
            log.info(f'delete worker:worker_id->{worker_id} success')
        except KeyError:
            log.warning(f'worker:worker_id->{worker_id} is not exist')
            return NotFound(msg=f'worker is not exist', errno=20004)

        try:
            while worker.is_alive():
                worker.terminate()
                time.sleep(0.5)
        except Exception as e:
            self.module_list[unique_id] = worker
            raise Conflict(f'worker stop error:{e}')
        return DeleteSuccess()

    def get_worker(self, worker_id):
        unique_id = f'worker_{worker_id}'
        worker = self.module_list.get(unique_id)
        if worker and worker.is_alive():
            return GetSuccess(data={'pid': worker.pid, 'status': 1})
        log.error(f'worker:worker_id->{worker_id} is not run')
        return GetSuccess(data={'pid': 0, 'status': -1})



from aiormq.types import DrainResult

from .pool import Pool


class ChannelSession(object):
    """
    提供和handler部分接口，入参返回完全一样
    使用此方法接收消息时，默认是手动ack模式
    """

    def __init__(self, pool: Pool):
        """非线程安全"""
        self.__pool = pool
        self.__channel = None

    async def __aenter__(self):
        self.__channel = await self.__pool._get()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.__channel is not None:
            self.__pool.put(self.__channel)

    # async def send_msg(self, msg: str, exchange: str, routing_key: str,
    #                    priority: int) -> typing.Optional[ConfirmationFrameType]:
    #     """
    #     发送消息 -> 默认持久化
    #     :param msg:
    #     :param exchange: 交换机名称
    #     :param routing_key: 路由关键字
    #     :param priority: 优先级 1-5 超过5将失效
    #     :return:
    #     """
    #     properties = spec.Basic.Properties(delivery_mode=2, priority=priority)
    #     return await self._do_query(
    #         func_name='basic_publish',
    #         exchange=exchange,
    #         routing_key=routing_key,
    #         body=msg.encode(),
    #         properties=properties)

    async def recv_msg(self, queue: str = '') -> tuple:
        msg = await self._do_query(func_name='basic_get', queue=queue, no_ack=False)
        if msg:
            method, properties, body, _ = msg
            if method is not None and isinstance(body, bytes):
                return method.delivery_tag, body.decode()
        return None, None

    async def _do_query(self, func_name, **kwargs):
        return await getattr(self.__channel, func_name)(**kwargs)

    # report ACK
    def ack(self, delivery_tag) -> DrainResult:
        return self._do_ack('basic_ack', delivery_tag)

    def nack(self, delivery_tag) -> DrainResult:
        return self._do_ack('basic_nack', delivery_tag)

    def _do_ack(self, func_name, delivery_tag):
        return getattr(self.__channel, func_name)(delivery_tag=delivery_tag)



import datetime

from bspider.core.broker import RabbitMQBroker
from bspider.utils.logger import LoggerPool
from bspider.utils.sign import Sign
from bspider.config.default_settings import EXCHANGE_NAME


class AsyncScheduler(object):

    def __init__(self, project_id: int, project_name: str, rate: int, sign: Sign, log_fn: str):
        self.sign = sign
        self.project_name = project_name
        self.project_id = project_id

        self.log = LoggerPool().get_logger(key=f'scheduler->{self.project_id}', fn=log_fn, module='scheduler', project=self.project_name)

        # 上一次分钟数
        self.__pre_loop_sign = None
        self.__scheduler_count = 0
        self.__downloader_queue = f'{EXCHANGE_NAME[1]}_{self.project_id}'
        self.rate = rate

    async def scheduler(self, broker: RabbitMQBroker):
        now = datetime.datetime.now()
        cur_loop_sign = 'scheduler-' + now.strftime('%H%M')
        # 表示新的调度周期
        if cur_loop_sign != self.__pre_loop_sign:
            self.__pre_loop_sign = cur_loop_sign
            self.__scheduler_count = 0
        self.log.debug(f'scheduler->{self.project_id} schedule cycle {cur_loop_sign} start!')
        cur_slice = int(int(now.strftime('%S')) // (60 / 12) + 1)
        self.log.debug(f'cur_slice: {cur_slice}')

        if not (self.rate < 1 or await self.__is_full_queue(broker, self.__downloader_queue, self.rate)):
            # 本次调度数量
            rate_slice = self.rate / 12
            self.log.debug(f'rate_slice: {rate_slice} real plan schedule num {int(rate_slice * 4)}')
            for i in range(int(rate_slice * 4)):
                # 获取项目本分钟内已经推送的数量,和当前时间段的需要推送量比较
                if int(rate_slice * cur_slice) > self.__scheduler_count:
                    if await broker.schedule_task(self.project_id):
                        self.__scheduler_count += 1
                else:
                    break
        self.log.debug(f'scheduler->{self.project_id} schedule cycle {cur_loop_sign} finish!')

    async def __is_full_queue(self, broker, queue_name, rate):
        """检查任务的下载队列是否阻塞"""
        msg_count = await broker.mq_client.get_queue_message_count(queue_name)
        threshold = rate // 12
        if threshold < 2:
            threshold = 1
        return msg_count > threshold
    #
    # async def __is_empty_queue(self, queue_name):
    #     count = await self.__broker.mq_handler.get_queue_message_count(queue_name)
    #     return count == 0

    def __repr__(self):
        return f'<AsyncScheduler->{self.project_name}:{self.sign}>'



from .agent_cache import AgentCache
from .base_manager import BaseManager
from .base_monitor import BaseMonitor
from .project import Project
from .broker import RabbitMQBroker
from .custom_module import BaseCustomModule

__all__ = [
    'AgentCache',
    'BaseManager',
    'RabbitMQBroker',
    'BaseMonitor',
    'Project',
    'BaseCustomModule'
]


"""
    返回一个包含配置的默认字典
"""
import os

from bspider.config import default_settings
from bspider.utils import singleton
from bspider.utils.conf import init_platform_env, PLATFORM_PATH_ENV
from bspider.utils.importer import import_module_by_path


@init_platform_env
@singleton
class FrameSettings(object):

    def __init__(self):
        """FrameSettings('${platform_name}.config.frame_settings')"""
        self.__settings = dict()

        for key in dir(default_settings):
            if key.isupper():
                self.__settings[key] = getattr(default_settings, key)
        # 优化错误处理
        module = import_module_by_path('frame_settings',
                                       os.path.join(os.environ[PLATFORM_PATH_ENV], 'config', 'frame_settings.py'))
        for key in dir(module):
            if key.isupper():
                self.__settings[key] = getattr(module, key)

    def get(self, key, default=None):
        if default:
            return self.__settings.get(key, default)
        else:
            return self.__settings[key]

    def __getitem__(self, item):
        return self.__settings[item]


if __name__ == '__main__':
    print(FrameSettings())



"""
    封装查看节点信息操作为api
"""
from flask import Blueprint

from bspider.agent.service.node import NodeService
from bspider.core.api import auth
from .validators.worker_forms import RegisterForm

node = Blueprint('node_bp', __name__)

node_service = NodeService()

@node.route('/node', methods=['GET'])
def status():
    """node-status 路由无需保护"""
    return node_service.node_status()

@node.route('/worker', methods=['POST'])
@auth.login_required
def start_worker():
    form = RegisterForm()
    return node_service.start_worker(**form.to_dict())

@node.route('/worker/<int:worker_id>', methods=['DELETE'])
@auth.login_required
def stop_worker(worker_id):
    return node_service.stop_worker(worker_id)

@node.route('/worker/<int:worker_id>', methods=['GET'])
@auth.login_required
def get_worker(worker_id):
    return node_service.get_worker(worker_id)



from flask import Blueprint

from bspider.core.api import auth
from bspider.master.service.chart import ChartService

chart = Blueprint('chart_bp', __name__)

chart_service = ChartService()


@chart.route('/downloader', methods=['GET'])
@auth.login_required
def downloader_pv():
    return chart_service.downloader_pv()


@chart.route('/downloader/<int:project_id>', methods=['GET'])
@auth.login_required
def downloader_pv_by_project(project_id):
    return chart_service.downloader_pv(project_id)


@chart.route('/parser', methods=['GET'])
@auth.login_required
def parser_pv():
    return chart_service.parser_pv()


@chart.route('/parser/<int:project_id>', methods=['GET'])
@auth.login_required
def parser_pv_by_project(project_id):
    return chart_service.parser_pv(project_id)


@chart.route('/code-type-detail', methods=['GET'])
@auth.login_required
def code_type_detail():
    return chart_service.get_code_type_detail()


@chart.route('/node-detail/<string:node_ip>', methods=['GET'])
@auth.login_required
def node_pv(node_ip):
    return chart_service.node_pv(node_ip)



"""
mysql 的异步封装
"""
import asyncio
import hashlib

import aiomysql
from aiomysql import Pool, DictCursor

from bspider.utils import singleton
from bspider.utils.tools import coroutine_result


@singleton
class AioMysqlPoolFactory(object):
    """
    mysql 异步连接池的连接池 不支持事务
    """

    def __init__(self):
        self.__pools = dict()

    def __del__(self):
        for key, pool in self.__pools.items():
            pool.close()

    @staticmethod
    def __make_hash(mysql_config):
        hashstr = ','.join([f'{k}:{v}' for k, v in mysql_config.items()])
        hashcode = hashlib.md5(hashstr.encode('utf8')).hexdigest()
        return hashcode

    def get_pool(self, mysql_config) -> Pool:
        hashcode = self.__make_hash(mysql_config)
        if hashcode in self.__pools:
            return self.__pools[hashcode]
        else:
            pool = coroutine_result(aiomysql.create_pool(
                host=mysql_config['host'],
                port=mysql_config['port'],
                user=mysql_config['user'],
                password=mysql_config['password'],
                db=mysql_config['db'],
                charset=mysql_config['charset'],
                cursorclass=DictCursor
            ))
            self.__pools[hashcode] = pool
            return pool


class AioMysqlClient(object):
    """封装了常用操作的mysql句柄，重试默认3次"""

    def __init__(self, config):
        self.__pool = AioMysqlPoolFactory().get_pool(config)

    @classmethod
    def from_settings(cls, settings):
        """通过配置创建AioMysqlHandler对象"""
        return cls(settings)

    async def select(self, sql: str, values: tuple = None,
                     retry_times: int = 0, retry_interval: int = 1) -> list:
        """
        提供重试的查询方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 查询结果
        """
        return await self._do_query(self._select, sql, values=values,
                                    retry_times=retry_times,
                                    retry_interval=retry_interval)

    async def insert(self, sql: str, values: tuple = None,
                     retry_times: int = 0, retry_interval: int = 1) -> int:
        """
        提供重试的插入方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 插入结果
        """
        return await self._do_query(self._insert, sql, values=values,
                                    retry_times=retry_times,
                                    retry_interval=retry_interval)

    async def update(self, sql: str, values: tuple = None,
                     retry_times: int = 0, retry_interval: int = 1) -> int:
        """
        提供重试的更新方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 更新结果
        """
        return await self._do_query(self._update, sql, values=values,
                                    retry_times=retry_times,
                                    retry_interval=retry_interval)

    async def delete(self, sql: str, values: tuple = None,
                     retry_times: int = 0, retry_interval: int = 1) -> int:
        """
        提供重试的删除方法方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 删除行数
        """
        return await self._do_query(self._delete, sql, values=values,
                                    retry_times=retry_times,
                                    retry_interval=retry_interval)

    async def query(self, sql: str, values: tuple = None,
                    retry_times: int = 0, retry_interval: int = 1):
        """
        query
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: T
        """
        return await self._do_query(self._query, sql, values=values,
                                    retry_times=retry_times,
                                    retry_interval=retry_interval)

    async def _do_query(self, func, sql: str, values: tuple = None,
                        retry_times: int = 0, retry_interval: int = 1):
        loop_times = retry_times if retry_times > 0 else 3
        while loop_times > 0:
            try:
                return await func(sql, values)
            except aiomysql.Error as e:
                errid, errmsg = e.args
                if errid != 2014:  # 不是mysql连接失败错误
                    raise e
            await asyncio.sleep(retry_interval)
        raise aiomysql.MySQLError('mysql connection error')

    async def _query(self, sql: str, values: tuple = None) -> int:
        async with self.__pool.acquire() as conn:
            async with conn.cursor() as cur:
                if values is None:
                    result = await cur.execute(sql)
                else:
                    result = await cur.execute(sql, values)
                await conn.commit()
                return result

    async def _select(self, sql: str, values: tuple = None) -> list:
        async with self.__pool.acquire() as conn:
            async with conn.cursor() as cur:
                if values is None:
                    result_count = await cur.execute(sql)
                else:
                    result_count = await cur.execute(sql, values)
                if result_count > 0:
                    return await cur.fetchall()
                else:
                    return []

    async def _insert(self, sql: str, values: tuple = None) -> int:
        return await self._query(sql, values)

    async def _update(self, sql: str, values: tuple = None) -> int:
        return await self._query(sql, values)

    async def _delete(self, sql: str, values: tuple = None) -> int:
        return await self._query(sql, values)





class DBSession(object):

    def __init__(self, pool, select_func):
        """得到一个线程安全连接"""
        self.__conn = pool.dedicated_connection()
        self.__cursor = self.__conn.cursor()
        self.__select = select_func

    def insert(self, sql: str, values: tuple = None, lastrowid: bool = False) -> int:
        return self._query(sql, values, lastrowid)

    def update(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def delete(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def select(self, sql: str, values: tuple = None,
               retry_times: int = 0, retry_interval: int = 1) -> list:
        return self.__select(sql,
                             values=values,
                             retry_times=retry_times,
                             retry_interval=retry_interval)

    def query(self, sql: str, values: tuple = None, lastrowid: bool = False):
        """
        query func
        :param sql: sql
        :param values: sql中需要初始化的值
        :param lastrowid:
        :return:
        """
        return self._query(sql, values=values, lastrowid=lastrowid)

    def _query(self, sql: str, values: tuple = None, lastrowid: bool = False):
        if values is None:
            result = self.__cursor.execute(sql)
        else:
            result = self.__cursor.execute(sql, values)

        if lastrowid:
            return self.__cursor.lastrowid
        else:
            return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val or exc_tb:
            self.__cursor.close()
            self.__conn.close()
            return False
        else:
            self.__cursor.close()
            self.__conn.commit()
            self.__conn.close()
            return True



"""
rabbitMQ 的混合类，提供rabbitMQ 推拉消息的封装
"""
import json
import logging

from bspider.config import FrameSettings
from bspider.config.default_settings import EXCHANGE_NAME
from bspider.http import Request, Response
from bspider.utils.rabbitMQ import AioRabbitMQClient


class RabbitMQBroker(object):
    # 要使用ID作为routing_key 必须要转为字符串否则无法绑定

    def __init__(self, log: logging.Logger, max_channel_size: int):
        self.frame_settings = FrameSettings()
        self.mq_client = AioRabbitMQClient(self.frame_settings['RABBITMQ_CONFIG'], max_channel_size)
        self.log = log
        self.log.debug('rabbitMQ broker init success')

    async def set_request(self, request: Request, project_id: int) -> bool:
        # 这里dump方法使用了浅拷贝，会影响一部分性能
        data = json.dumps(request.dumps())
        await self.mq_client.send_msg(data, EXCHANGE_NAME[0], str(project_id), request.priority)
        self.log.info(f'success set a new Request: {request}')
        return True

    async def set_response(self, response: Response, project_id: int) -> bool:
        """将解析结果发送到不同的exchange"""
        # 这里dump方法使用了浅拷贝，会影响一部分性能
        data = json.dumps(response.dumps())
        await self.mq_client.send_msg(data, EXCHANGE_NAME[2], str(project_id), response.request.priority)
        self.log.debug(f'success set a new Response: {response}')
        return True
    
    async def schedule_task(self, project_id: int) -> bool:
        """调度抓取任务到下载队列"""
        self.log.debug(f'start to schedule task')
        queue_name = '{}_{}'.format(EXCHANGE_NAME[0], project_id)

        async with self.mq_client.session() as session:
            msg_id, data = await session.recv_msg(queue_name)
            if msg_id is not None:
                request = Request.loads(json.loads(data))
                if await self.mq_client.send_msg(data, EXCHANGE_NAME[1], str(project_id), priority=request.priority):
                    self.log.info(f'send a new task to request queue=> project_id:{project_id}, body:{data}')
                    session.ack(msg_id)
                    return True
                else:
                    self.log.warning(f'send a new task fail sign->{request.sign}')
                    session.nack(msg_id)
        return False



import asyncio
import json
import sys
import traceback

from bspider.core import BaseManager
from bspider.http import Response, ERROR_RESPONSE
from bspider.config.default_settings import EXCHANGE_NAME

from .parser_monitor import ParserMonitor


def run_parser(unique_sign, coro_num=3):
    dm = ParserManager(unique_sign, ParserMonitor, coro_num)
    dm.run()


class ParserManager(BaseManager):
    manager_type = 'parser'
    exchange = EXCHANGE_NAME[2]

    async def do_work(self):
        try:
            while True:
                parser = self.monitor.projects.get(self.monitor.choice_project())
                if parser is None:
                    # 防止协程抢占无法轮换
                    await asyncio.sleep(1)
                    continue
                e_msg = None
                response = ERROR_RESPONSE

                async with self.broker.mq_client.session() as session:
                    msg_id, data = await session.recv_msg(f'{self.exchange}_{parser.project_id}')
                    if msg_id:
                        try:
                            response = Response.loads(json.loads(data))
                            self.log.info(f'success get a new Response: {response}')
                            requests = await parser.parse(response)
                        except Exception as e:
                            tp, msg, tb = sys.exc_info()
                            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
                            e.with_traceback(tb)
                            self.log.exception(e)
                        session.ack(msg_id)

                if msg_id:
                    if e_msg:
                        await self._save_error_result(response, parser.project_name, parser.project_id, e_msg)
                        self.log.warning('parser failed: project:project_id->{} project_name->{} sign: {}'.format(
                            parser.project_id, parser.project_name, response.sign))
                    else:
                        while len(requests):
                            # 发送request到待下载队列
                            await self.broker.set_request(requests.pop(), parser.project_id)
                        await self._save_success_result(response, parser.project_name, parser.project_id)
                        self.log.info('project:project_id->{} project_name->{} complete parser: {}'.format(
                            parser.project_id, parser.project_name, response.url))
        except Exception:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            self.log.error(f'coro error:{e_msg}')


if __name__ == '__main__':
    run_parser('parser-01')



from flask import g
from flask_httpauth import HTTPTokenAuth
from .token import verify_token

auth = HTTPTokenAuth()


@auth.verify_token
def verify_ctx_token(x_token: str) -> bool:
    user = verify_token(x_token)
    if user:
        g.user = user
        return True
    return False



from flask import Blueprint

from bspider.core.api import auth
from bspider.master.service.rabbitmq import RabbitMQService

rabbitmq = Blueprint('rabbitmq_bp', __name__)

rabbitmq_service = RabbitMQService()


@rabbitmq.route('/project/<int:project_id>', methods=['GET'])
@auth.login_required
def project_queue_info(project_id):
    """project相关队列的详细信息"""
    return rabbitmq_service.get_project_queue_info(project_id)


@rabbitmq.route('/project/purge/<int:project_id>', methods=['DELETE'])
@auth.login_required
def purge_project_queue(project_id):
    """清空待下载链接"""
    return rabbitmq_service.purge_project_queue(project_id)



import asyncio

from bspider.config import FrameSettings
from bspider.core import BaseManager
from .scheduler_monitor import SchedulerMonitor


def run_scheduler(coro_num=1):
    """A factory to make a download process"""
    coro_num = 1 if coro_num != 1 else coro_num
    dm = SchedulerManager(FrameSettings().get('SCHEDULER_SIGN', 'default'), SchedulerMonitor, coro_num)
    dm.run()


class SchedulerManager(BaseManager):
    manager_type = 'scheduler'

    async def do_work(self):
        while True:
            self.log.debug('A new schedule start!')
            for project_id, scheduler in self.monitor.projects.items():
                await scheduler.scheduler(self.broker)
                self.log.debug(f'schedule project:project_id=>{project_id}!')
            self.log.info(f'A new schedule complete ! project_count=>{len(self.monitor.projects)}')
            await asyncio.sleep(1)


if __name__ == '__main__':
    run_scheduler()



import os
import sys
import traceback

from flask import Flask, current_app
from flask_cors import CORS
from werkzeug.exceptions import HTTPException

import bspider
from bspider.config import FrameSettings
from bspider.core.api import APIException
from bspider.master import log
from bspider.master.controller.chart import chart
from bspider.master.controller.cron import cron
from bspider.master.controller.code import code
from bspider.master.controller.project import project
from bspider.master.controller.node import node
from bspider.master.controller.rabbitmq import rabbitmq
from bspider.master.controller.user import user
from bspider.master.controller.tools import tools


def create_app():
    app = Flask(
        __name__,
        template_folder=os.path.join(bspider.__path__[0], 'master', 'ui'),
        static_folder=os.path.join(bspider.__path__[0], 'master', 'ui', 'static')
    )
    CORS(app)
    frame_settings = FrameSettings()
    app.secret_key = frame_settings['WEB_SECRET_KEY']
    # 用户管理模块
    app.register_blueprint(user)
    # project管理模块
    app.register_blueprint(project)
    # 节点管理模块
    app.register_blueprint(node)
    # code管理模块
    app.register_blueprint(code)
    # cron_job
    app.register_blueprint(cron)
    # tools
    app.register_blueprint(tools, url_prefix="/tools")
    # rabbitmq
    app.register_blueprint(rabbitmq, url_prefix="/rabbitmq")
    # chart
    app.register_blueprint(chart, url_prefix="/chart")


    @app.errorhandler(Exception)
    def framework_error(error):
        if isinstance(error, APIException):
            return error
        if isinstance(error, HTTPException):
            return APIException(error.code, error.description, 10000)
        tp, msg, tb = sys.exc_info()
        e_msg = ''.join(traceback.format_exception(tp, msg, tb))
        log.error(e_msg)
        return APIException(msg=str(error))

    @app.route('/favicon.ico', methods=['GET'])
    def get_fav():
        return current_app.send_static_file('favicon.ico')

    app.logger.addHandler(log)

    return app



"""
** 检查是否有supervisor.pid文件确认supervisor是否启动，启动则调用rpc接口启动进程 否则：初始化supervisor.conf 启动进程
1. web service
2. 调度器
3. 定时任务管理
"""
import os
from shutil import copy2

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_PATH_ENV, PLATFORM_NAME_ENV
from bspider.utils.exceptions import UsageError
from bspider.utils.template import render_templatefile


class Command(BSpiderCommand):

    def syntax(self):
        return "<op:start|stop>"

    def short_desc(self):
        return "Run BSpider as a agent node"

    def long_desc(self):
        return """Run BSpider by supervisor with one process:
        agent: a web server to manager bspider worker node (by gunicorn and gevent).
        """

    def init_supervisor(self):
        """查看supervisor是否已经启动"""
        platform_path = os.environ[PLATFORM_PATH_ENV]
        platform_name = os.environ[PLATFORM_NAME_ENV]

        tplfile = os.path.join(self.templates_dir, 'tools_cfg', 'agent_gunicorn.py.tmpl')
        copy2(tplfile, os.path.join(platform_path, '.cache', 'agent_gunicorn.py.tmpl'))
        render_templatefile(os.path.join(platform_path, '.cache', 'agent_gunicorn.py.tmpl'),
                            agent_port=self.frame_settings['AGENT']['port'],
                            agent_ip=self.frame_settings['AGENT']['ip'],
                            log_level=self.frame_settings['LOGGER_LEVEL'].lower(),
                            platform_name=platform_name,
                            platform_path=platform_path)

        if os.path.exists(os.path.join(platform_path, '.cache', 'supervisord.pid')):
            return True
        tplfile = os.path.join(self.templates_dir, 'tools_cfg', 'supervisor.conf.tmpl')
        config_path = os.path.join(platform_path, '.cache', 'supervisor.conf')
        copy2(tplfile, config_path)
        render_templatefile(config_path,
                            platform_path=platform_path,
                            bin_path=os.path.join(platform_path, 'bin'),
                            master_ip=self.frame_settings['MASTER']['ip'],
                            supervisor_rpc_port=self.frame_settings['SUPERVISOR_RPC']['port'],
                            supervisor_rpc_username=self.frame_settings['SUPERVISOR_RPC']['username'],
                            supervisor_rpc_password=self.frame_settings['SUPERVISOR_RPC']['password'])

        cmd = 'supervisord -c {}'.format(os.path.join(platform_path, '.cache', 'supervisor.conf'))
        print(os.popen(cmd).read().strip())
        return True

    def run(self, args, opts):
        if len(args) != 1:
            raise UsageError('args error')
        self.init_supervisor()
        op = args[0]
        if op not in ('start', 'stop'):
            raise UsageError('unknow op: %s' % (op))
        rpc_socket = os.path.join(os.environ[PLATFORM_PATH_ENV], '.cache', 'supervisor.conf')
        cmd = f'supervisorctl -c {rpc_socket} {op} agent'
        print(cmd)
        print(os.popen(cmd).read().strip())
        print(f'A new BSpider agent node {op} success')


if __name__ == '__main__':
    Command().run(['hello_project'])



"""
    封装对cache中code信息的操作为api
"""
from flask import Blueprint

from bspider.core.api import auth
from .validators.code_form import AddForm, UpdateForm
from bspider.agent.service.code import CodeService

code = Blueprint('code_bp', __name__)
code_service = CodeService()


@code.route('/code', methods=['POST'])
@auth.login_required
def add_code():
    form = AddForm()
    return code_service.add_code(**form.to_dict())


@code.route('/code/<int:code_id>', methods=['PATCH'])
@auth.login_required
def update_code(code_id):
    form = UpdateForm()
    return code_service.update_code(code_id, form.to_dict())


@code.route('/code/<int:code_id>', methods=['DELETE'])
@auth.login_required
def delete_code(code_id):
    return code_service.delete_code(code_id)


@code.route('/code', methods=['GET'])
@auth.login_required
def get_codes():
    return code_service.get_codes()


@code.route('/code/<int:code_id>', methods=['GET'])
def get_code(code_id):
    return code_service.get_code(code_id)



from pymysql import IntegrityError

from bspider.core.api import BaseService, Conflict, NotFound, PostSuccess, PatchSuccess, DeleteSuccess, \
    GetSuccess, ParameterException
from bspider.core.api import AgentMixIn
from bspider.master.server import log
from bspider.master.service.impl.code_impl import CodeImpl


class CodeService(BaseService, AgentMixIn):

    def __init__(self):
        self.impl = CodeImpl()

    def add(self, content, editor):
        name, code_type, description, code = content
        try:
            with self.impl.mysql_client.session() as session:
                data = {
                    'name': name,
                    'description': description,
                    'type': code_type,
                    'content': code,
                    'editor': editor
                }
                code_id = session.insert(*self.impl.add_code(data=data))

                sign, result = self.op_add_code(self.impl.get_nodes(), {'code_id': code_id, 'content': code})
                if not sign:
                    log.warning(f'not all node add cide:{name} =>{result}')
                    raise Conflict(msg=f'not all node add code:{name}', data=result, errno=40006)
            log.info(f'add code success:{name}-{editor}')
            return PostSuccess(msg='add code success', data={'code_id': code_id})
        except IntegrityError:
            log.error(f'add code failed:{name}-{editor} code is already exist')
            return Conflict(msg='code is already exist', errno=40002)

    def update(self, code_id, content, editor):
        infos = self.impl.get_code(code_id)
        if not len(infos):
            return NotFound(msg='code is not exist', errno=40001)

        project = list()
        for info in infos:
            project.append(str(info['project_id']))
        infos[0].pop('project_id')
        infos[0].pop('project_name')
        infos[0]['project'] = project
        info = infos[0]
        update_info = dict()

        with self.impl.mysql_client.session() as session:
            name, code_type, description, code = content
            if name != info['name']:
                update_info['name'] = name

            if editor != info['editor']:
                update_info['editor'] = editor

            if code_type != info['type']:
                update_info['type'] = code_type

            if description != info['description']:
                update_info['description'] = description

            if code != info['content']:
                sign, result = self.op_update_code(self.impl.get_nodes(), code_id,
                                                   {'content': code, 'project': ','.join(project)})
                if not sign:
                    log.warning(f'code:code_id->:{code_id} update exception')
                    raise Conflict(msg=f'not all code:code_id->:{code_id} was update', data=result, errno=40006)
                update_info['content'] = code

            session.update(*self.impl.update_code(code_id, update_info))
        log.info(f'update code success: {update_info}')
        return PatchSuccess(msg='update code success')

    def delete(self, code_id):
        project_list = self.impl.get_project_by_code_id(code_id)
        if len(project_list):
            log.error(f'delete code:{code_id} failed: can\'t delete in use code')
            return Conflict(msg='can\'t delete in use code', data=project_list, errno=40004)
        else:
            with self.impl.mysql_client.session() as session:
                sign, result = self.op_delete_code(self.impl.get_nodes(), code_id)
                if not sign:
                    log.warning(f'code:code_id->:{code_id} delete exec')
                    raise Conflict(msg=f'not all code:code_id->:{code_id} was delete', data=result, errno=40006)
                session.delete(*self.impl.delete_code(code_id))
            log.info(f'success delete code:{code_id}')
            return DeleteSuccess()

    def get_code(self, code_id):
        infos = self.impl.get_code(code_id)
        if len(infos):
            project = []
            sign = set()
            for info in infos:
                if info['project_id'] and info['project_id'] not in sign:
                    project.append({'id': info['project_id'], 'name': info['project_name']})
                    sign.add(info['project_id'])
            infos[0].pop('project_id')
            infos[0].pop('project_name')
            infos[0]['project'] = project
            return GetSuccess(msg='get code success', data=infos[0])
        else:
            return NotFound(msg='code is not exist', errno=40001)

    def get_codes(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_codes(page, limit, search, sort)

        for info in infos:
            self.datetime_to_str(info)

        return GetSuccess(
            msg='get user list success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })



import os
from os.path import join, exists

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_PATH_ENV
from bspider.utils.exceptions import UsageError
from bspider.utils.template import render_templatefile
from bspider.utils.tools import module_name2class_name

TEMPLATES_TO_RENDER = (
    ('__init__.py.tmpl',),
    ('extractor.py.tmpl',),
    ('settings-development.yml.tmpl',),
    ('settings-production.yml.tmpl',),
)


class Command(BSpiderCommand):

    def syntax(self):
        return "<spider_name>"

    def short_desc(self):
        return "Create new spider in current platform"

    def run(self, args, opts):
        if len(args) not in (1, 2):
            raise UsageError('args error')

        project_name = args[0]
        project_dir = join(os.environ[PLATFORM_PATH_ENV], 'projects', project_name)
        if exists(project_dir):
            self.exitcode = 1
            print(f'Error: {project_name} already exists in this platform')
            return

        self._copytree(join(self.templates_dir, 'spiders'), project_dir)
        for paths in TEMPLATES_TO_RENDER:
            tplfile = join(project_dir, *paths)
            render_templatefile(tplfile, project_name=project_name, ProjectName=module_name2class_name(project_name))

        print(f"New spider '{project_name}' init success")
        print(f"    template directory:\n    '{self.templates_dir}'")
        print("You can development your spider in:")
        print(f"    cd {project_dir}")


if __name__ == '__main__':
    Command().run(['hello_project'])



import logging
from asyncio import iscoroutinefunction

from bspider.core import Project


class BaseCustomModule(object):

    def __init__(self, project: Project, settings: dict, log: logging.Logger):
        """
        构造函数，提供当前project对象，和当前模块的
        :param project: project对象
        :param settings: 当前模块的参数
        :param log: 日志句柄
        """
        self.project = project
        self.settings = settings
        self.log = log

    async def _exec(self, func_name, *args):
        """用以支持协程和非协程编程方式"""
        func = getattr(self, func_name)
        if iscoroutinefunction(func):
            return await func(*args)
        else:
            return func(*args)


"""
整个定时任务模块基于APSchedule 定制化开发而来
解决 gunicorn + Flask 启动多次实例化APSchedule导致定时任务多次执行的bug
这里采用单独进程处理任务调度
"""
import sys
import time
import traceback

import pytz
from apscheduler.executors.pool import ThreadPoolExecutor
from datetime import datetime

from apscheduler.triggers.cron import CronTrigger

from bspider.config import FrameSettings
from bspider.bcron import do
from bspider.bcron.jobstore import MySQLJobStore
from bspider.bcron.scheduler import MySQLScheduler
from bspider.utils.logger import LoggerPool
from bspider.utils.database import MysqlClient

def run_bcorn():
    """A factory to make a download process"""
    dm = BCronManager()
    dm.run()


class BCronManager(object):

    def __init__(self):
        self.frame_settings = FrameSettings()
        self.log = LoggerPool().get_logger(key='bcorn_manager', fn='bcorn', module='bcorn')
        self.mysql_client = MysqlClient.from_settings(self.frame_settings['WEB_STUDIO_DB'])
        self.table_name = self.frame_settings['CRON_JOB_STORE_TABLE']
        self.tz = pytz.timezone(self.frame_settings['TIMEZONE'])
        self.interval = self.frame_settings['CRON_JOB_REFRESH_TIME']

        store = MySQLJobStore(self.mysql_client, self.tz, self.log, self.table_name)

        executors = {
            'thread_pool': ThreadPoolExecutor(self.frame_settings['CRON_JOB_THREAD_NUM'])
            # 'processpool': ProcessPoolExecutor(frame_settings.CRON_JOB_PROCESS_NUM)
        }
        self.log.debug(
            'success make executors thread num:{}'.format(self.frame_settings['CRON_JOB_THREAD_NUM']))

        self.scheduler = MySQLScheduler(
            jobstores={'default': store},
            executors=executors,
            timezone=self.tz
        )
        # 之前检查的时间
        self.pre_time = self.get_now()

    def get_now(self):
        return datetime.now(self.tz).strftime("%Y-%m-%d %H:%M:%S.%f")

    def check_change(self):

        sql = f'select `id`, `project_id`, `code_id`, `trigger`, `trigger_type`, `type`, ' \
              f'`description`, `next_run_time`, `status` from {self.table_name} ' \
              f'where update_time >= date_sub(now(), interval {self.interval} second)         ' \
              f'and `status` != 0;'
        infos = self.mysql_client.select(sql)

        for info in infos:
            if info['status'] == 1:
                self.real_add_job(info)
            elif info['status'] == 2:
                self.real_update_job(info)
            elif info['status'] == 3:
                self.real_remove_job(info['id'])

    def real_remove_job(self, job_id):
        self.scheduler.remove_job(job_id)
        self.log.info('remove job success!')

    def real_update_job(self, info):
        info.pop('trigger_type')
        self.scheduler.modify_job(
            job_id=info['id'],
            name='{project_id}-{code_id}'.format(**info),
            status=0,
            trigger=CronTrigger.from_crontab(info['trigger']),
            cron_type=info['type'],
            description=info['description']
        )

    def real_add_job(self, info):
        name = '{project_id}-{code_id}'.format(**info)
        cron = CronTrigger.from_crontab(info['trigger'])
        now = datetime.now(self.tz)
        next_run_time = cron.get_next_fire_time(None, now)
        try:
            self.scheduler.add_job(
                do, cron, name=name,
                kwargs={'project_id': info['project_id'], 'code_id': info['code_id'], 'cron_type': info['type']},
                description=info['description'],
                next_run_time=next_run_time,
                id=info['id'],
                cron_type=info['type']
            )
        except Exception as e:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            self.log.exception(e_msg)
            self.log.error(f'real add job failed: {name} {e} {info}')
            return
        self.log.info(f'success add job {name}, it run at {next_run_time}')

    def run(self):
        self.log.debug('success init scheduler!')
        self.scheduler.start()
        self.log.info('success start MySQLScheduler')
        # 每三秒检查一下定时任务是否有变化
        while True:
            self.check_change()
            time.sleep(self.interval)


if __name__ == '__main__':
    BCM = BCronManager()
    BCM.run()


"""
2019-08-29 增加callback 参数处理 - 需要单独抽象出extractor类
将Request 对象发送到待下载队列
2020-03-18 增加SendRequestHelper
"""
from types import GeneratorType

from bspider.core import Project
from bspider.http import Response, Request
from bspider.utils.exceptions import ParserError
from bspider.utils.logger import LoggerPool
from bspider.utils.importer import import_module_by_code
from bspider.utils.sign import Sign


class SendRequestHelper(object):

    def process_item(self, item):
        if isinstance(item, Request):
            yield item

    async def _exec(self, func_name, *args):
        return getattr(self, func_name)(*args)


class AsyncParser(object):

    def __init__(self, project: Project, sign: Sign, log_fn: str):
        """传入下载器的配置文件"""
        self.sign = sign
        self.project_name = project.project_name
        self.project_id = project.project_id

        self.log = LoggerPool().get_logger(
            key=f'project_parser->{self.project_id}', fn=log_fn, module='parser', project=self.project_name)

        self.pipes = []
        for pipeline in project.parser_settings.pipeline:
            for cls, params in pipeline.items():
                cls_name, code = cls
                if isinstance(code, str):
                    mod = import_module_by_code(cls_name, code)
                else:
                    mod = code
                if mod:
                    if hasattr(mod, cls_name):
                        try:
                            self.pipes.append(
                                getattr(mod, cls_name)(project, {**project.global_settings, **params}, self.log))
                            self.log.info(f'success load: <{self.project_name}:{cls_name}>!')
                        except Exception as e:
                            raise ParserError(
                                '%s pipeline init failed: %s like: %s' % (self.project_name, cls_name, e))
                else:
                    msg = f'{self.project_name} pipeline init failed: {cls_name}'
                    raise ParserError(msg)
        # 增加默认的 SendRequestHelper
        self.pipes.append(SendRequestHelper())

    async def parse(self, response: Response) -> list:
        """
        :param response:
        :return:
        """
        all_items = [response]
        requests = []

        for index, pipeline in enumerate(self.pipes):
            cur_items = []
            for item in all_items:
                await self.__get_items(pipeline, item, cur_items, requests)
            all_items = cur_items

        return requests

    async def __get_items(self, pipeline, pre_item, cur_item, requests):
        """
        得到产生的item
        :param pipeline: 当前 pipline 对象
        :param pre_item: 上一个 pipline 产生的item
        :param cur_item: 这个 pipline 产生的item
        :return:
        """
        self.log.debug(f'{pipeline.__class__.__name__} executing process_response')
        temp_items = await pipeline._exec('process_item', pre_item)
        is_send_request_helper = isinstance(pipeline, SendRequestHelper)
        self.__exec_items(temp_items, requests, cur_item, is_send_request_helper)

    def __exec_items(self, items, requests: list, cur_item: list, is_send_request_helper: bool):
        if not isinstance(items, GeneratorType):
            if isinstance(items, Request):
                if is_send_request_helper:
                    requests.append(items)
            cur_item.append(items)
            return

        while True:
            try:
                self.__exec_items(next(items), requests, cur_item, is_send_request_helper)
            except StopIteration as e:
                self.__exec_items(e.value, requests, cur_item, is_send_request_helper)
                break



import os

from bspider.core.api import BaseImpl, json
from bspider.utils.conf import PLATFORM_PATH_ENV
from bspider.utils.database import SqlLite3Client


class AgentCache(object):

    def __init__(self):
        self.sqlite3_client = SqlLite3Client(os.path.join(os.environ[PLATFORM_PATH_ENV], '.cache', 'meta.db'))
        self.project_table = 'project_cache'
        self.project_weight_table = 'project_weight_cache'
        self.code_table = 'code_table'

    def initialization(self) -> (bool, str):
        """初始化bdb"""
        sqls = """
        DROP table if exists {project_table};
        CREATE TABLE `{project_table}` (
            `id` int(11) PRIMARY KEY NOT NULL,
            `name` varchar(50) NOT NULL,
            `rate` int(10) NOT NULL,
            `config` text NOT NULL,
            `status` int(2) NOT NULL,
            `timestamp` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP
        ); 
        CREATE UNIQUE INDEX `idx_name` on {project_table} (`name`);
        CREATE UNIQUE INDEX `idx_id` on {project_table} (`id`);
        DROP table if exists {project_weight_table};
        CREATE TABLE `{project_weight_table}` (
            `weight_id` int(11) PRIMARY KEY NOT NULL,
            `project_weight` text NOT NULL,
            `timestamp` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP
        );
        DROP table if exists {code_table};
        CREATE TABLE `{code_table}` (
            `id` int(11) PRIMARY KEY NOT NULL,
            `content` text NOT NULL,
            `timestamp` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP
        ); 
        """.format(
            code_table=self.code_table,
            project_weight_table=self.project_weight_table,
            project_table=self.project_table
        )
        for sql in sqls.split(';'):
            self.sqlite3_client.query(sql)
        return True, 'ok'

    def get_projects(self):
        sql = f'select `id`, `name`, `config`, `rate`, `timestamp` from {self.project_table} where `status`=1;'
        return [{'id': project_id, 'name': name, 'config': json.loads(config),
                 'rate': rate, 'timestamp': timestamp} for project_id, name, config, rate, timestamp in
                self.sqlite3_client.select(sql)]

    def get_project(self, project_id):
        sql = f'select `id`, `name`, `config`, `rate`, `timestamp`, `status` from project_cache where `id`={project_id};'
        return [{'id': project_id, 'name': name, 'config': json.loads(config),
                 'rate': rate, 'timestamp': timestamp, 'status': status} for
                project_id, name, config, rate, timestamp, status in
                self.sqlite3_client.select(sql)]

    def set_project(self, project_id: int, name: str, config: str, rate: int, status: int):
        sql = f"insert or replace into {self.project_table} values(?, ?, ?, ?, ?, datetime('now'))"
        return self.sqlite3_client.insert(sql, (project_id, name, rate, json.dumps(config), status))

    def delete_project(self, project_id: int):
        sql = f'delete from {self.project_table} where `id`={project_id}'
        return self.sqlite3_client.delete(sql)

    def update_project(self, project_id, data):
        fields, values = BaseImpl.make_fv(data)
        sql = f"update {self.project_table} set {fields} where `id` = {project_id};".replace('%s', '?')
        return self.sqlite3_client.update(sql, values)

    def get_codes(self):
        sql = f'select `id`, `content` from {self.code_table};'
        return [{'id': code_id, 'content': content} for code_id, content in self.sqlite3_client.select(sql)]

    def get_code(self, code_id):
        sql = f'select `id`, `content` from {self.code_table} where `id`={code_id};'
        return [{'id': code_id, 'content': content} for code_id, content in self.sqlite3_client.select(sql)]

    def set_code(self, code_id: int, content: str):
        sql = f"insert or replace into {self.code_table} values(?, ?, datetime('now'))"
        return self.sqlite3_client.insert(sql, (code_id, content))


    def delete_code(self, code_id: str):
        sql = f'delete from {self.code_table} where `id`={code_id}'
        return self.sqlite3_client.delete(sql)

    def update_code(self, code_id: int, data: dict , project: list):
        fields, values = BaseImpl.make_fv(data)
        sql = f"update {self.code_table} set {fields} where `id` = {code_id};".replace('%s', '?')
        self.sqlite3_client.update(sql, values)
        # 在更新代码的时候 更新project 的时间
        for project_id in project:
            sql = f"update {self.project_table} set `timestamp`= CURRENT_TIMESTAMP where `id` = {project_id}"
            self.sqlite3_client.update(sql)



import logging

from bspider.config import FrameSettings


class BaseOperation(object):

    def __init__(self, log: logging.Logger):
        self.frame_settings = FrameSettings()
        self.log = log


    def execute_task(self):
        raise NotImplementedError



from bspider.config.default_settings import EXCHANGE_NAME
from bspider.core import BaseMonitor, Project
from bspider.utils.sign import Sign

from .async_downloader import AsyncDownloader


class DownloaderMonitor(BaseMonitor):

    exchange = EXCHANGE_NAME[1]

    def get_work_obj(self, project: Project, sign: Sign):
        return AsyncDownloader(project, sign, self.log_fn)




import sqlite3

from bspider.utils import singleton


@singleton
class SqlLite3Client(object):

    def __init__(self, path):
        self.conn = sqlite3.connect(path)

    @staticmethod
    def _do_query(func, sql: str, values: tuple = None, lastrowid: bool = False):
        if lastrowid:
            return func(sql, values, lastrowid)
        else:
            return func(sql, values)

    def select(self, sql: str, values: tuple = None) -> list:
        """
        提供重试的查询方法
        :param sql: sql语句
        :param values: sql语句中的值
        :return: 查询结果
        """
        return self._do_query(self._select, sql, values=values)

    def insert(self, sql: str, values: tuple = None, lastrowid: bool = False) -> int:
        """
        提供重试的插入方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param lastrowid: 是否返回自增id
        :return: 插入结果
        """
        return self._do_query(self._insert, sql, values=values, lastrowid=lastrowid)

    def update(self, sql: str, values: tuple = None) -> int:
        """
        提供重试的更新方法
        :param sql: sql语句
        :param values: sql语句中的值
        :return: 更新结果
        """
        return self._do_query(self._update, sql, values=values)

    def delete(self, sql: str, values: tuple = None) -> int:
        """
        提供重试的删除方法方法
        :param sql: sql语句
        :param values: sql语句中的值
        :return: 删除行数
        """
        return self._do_query(self._delete, sql, values=values)

    def query(self, sql: str, values: tuple = None):
        """
        query
        :param sql: sql语句
        :param values: sql语句中的值
        :return: T
        """
        return self._do_query(self._query, sql, values=values)

    def _select(self, sql: str, values: tuple = None) -> list:
        cursor = None
        try:
            cursor = self.conn.cursor()
            if values is None:
                cursor.execute(sql)
            else:
                cursor.execute(sql, values)
            return cursor.fetchall()
        finally:
            if cursor:
                cursor.close()

    def _insert(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def _update(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def _delete(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def _query(self, sql: str, values: tuple = None, lastrowid: bool = False):
        cursor = None
        try:
            cursor = self.conn.cursor()
            if values is None:
                result = cursor.execute(sql)
            else:
                result = cursor.execute(sql, values)

            if lastrowid:
                return cursor.lastrowid
            else:
                return result
        finally:
            if cursor:
                cursor.close()
            self.conn.commit()



def singleton(cls):
    """
    单例类
    :param cls:
    :param args:
    :param kw:
    :return:
    """
    instances = {}
    def _singleton(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return _singleton


"""
生成用于启动agent服务的app对象，是整个agent服务的入口
"""
import sqlite3
import sys
import traceback

from flask import Flask
from werkzeug.exceptions import HTTPException

from bspider.agent import log
from bspider.core.api import APIException, MasterMixIn
from bspider.core import AgentCache
from bspider.config import FrameSettings

from bspider.agent.controller.project import project
from bspider.agent.controller.node import node, node_service
from bspider.agent.controller.code import code

from bspider.utils.exceptions import RemoteOPError
from bspider.utils.system import System


class CreateApp(MasterMixIn):

    def __init__(self):
        self.frame_settings = FrameSettings()

    def app(self):
        app = Flask(__name__)
        app.secret_key = self.frame_settings['WEB_SECRET_KEY']
        app.register_blueprint(node)
        app.register_blueprint(project)
        app.register_blueprint(code)

        @app.errorhandler(Exception)
        def framework_error(error):
            if isinstance(error, APIException):
                return error
            if isinstance(error, HTTPException):
                return APIException(error.code, error.description, 10000)
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            log.error(f'server exec:{e_msg}')
            return APIException()

        if self.__init_agent(**self.frame_settings['AGENT']):
            log.info('Agent:{name}->{ip}:{port} run success!'.format(**self.frame_settings['AGENT']))
            app.logger.addHandler(log)
            return app
        else:
            exit()

    def __init_agent(self, ip, name, description, port) -> bool:
        cache = AgentCache()
        system = System()
        try:
            # 初始化缓存、获取缓存数据
            cache.initialization()
            data = self.op_add_node({
                'ip': ip,
                'description': description,
                'name': name,
                'cpu_num': system.cpu_num,
                'mem_size': system.mem_size,
                'disk_size': system.disk_size,
                'port': port
            })
        except sqlite3.Error:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            log.error(f'AgentCache init failed exec:{e_msg}')
            return False
        except RemoteOPError:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            log.error(f'Master server exec:{e_msg}')
            return False

        log.debug(
            'start to load master\'s data: project->{} code->{}'.format(len(data['projects']), len(data['codes'])))
        try:
            # 向缓存中写入数据
            for code in data['codes']:
                cache.set_code(**code)

            for project in data['projects']:
                cache.set_project(**project)
        except Exception:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            log.error(f'AgentCache init failed exec:{e_msg}')
            return False

        log.info('cache init complete!! project->{} code->{}'.format(len(data['projects']), len(data['codes'])))
        log.debug('start to recover Agent\'s worker: count->{}'.format(len(data['workers'])))

        # 恢复节点下工作的worker
        for worker in data['workers']:
            if node_service.start_worker(worker['worker_id'], worker['type'], worker['coroutine_num']).errno != 0:
                log.error('worker init failed: {worker_id} {type} {coroutine_num}'.format(**worker))
                return False

        log.info(' Agent\'s worker recover complete !! count->{}'.format(len(data['workers'])))
        return True



import asyncio
import hashlib
import time
import typing

import aiormq

from aiormq.types import ConfirmationFrameType
from pamqp import specification as spec

from bspider.config.default_settings import QUEUE_ARG
from bspider.utils import singleton
from .session import ChannelSession
from .pool import Pool


@singleton
class RabbitMQPoolFactory(object):
    """rabbitMQ 连接池的连接池"""
    base_url = "amqp://{username}:{password}@{host}:{port}/{virtual_host}"

    def __init__(self):
        self.__pools = dict()

    @staticmethod
    def __make_hash(mq_config: dict) -> str:
        hashstr = ','.join([f'{k}:{v}' for k, v in mq_config.items()])
        hashcode = hashlib.md5(hashstr.encode('utf8')).hexdigest()
        return hashcode

    def get_pool(self, mq_config: dict, max_size: int) -> Pool:
        hashcode = self.__make_hash(mq_config)
        if hashcode in self.__pools:
            return self.__pools[hashcode]
        else:
            pool = Pool(
                self.get_channel,
                Pool(self.get_connection, mq_config, max_size=1),
                max_size=max_size,
            )
            self.__pools[hashcode] = pool
            return pool

    async def get_connection(self, mq_config: dict) -> aiormq.Connection:
        return await aiormq.connect(self.base_url.format(**mq_config))

    @staticmethod
    async def get_channel(conn_pool: Pool) -> aiormq.Channel:
        async with conn_pool.acquire() as connection:
            return await connection.channel()


class AioRabbitMQClient(object):

    def __init__(self, config: dict, max_size: int = 3):
        self.__pool = RabbitMQPoolFactory().get_pool(mq_config=config, max_size=max_size)
        self.queue_arg = QUEUE_ARG

    @classmethod
    def from_settings(cls, settings: dict):
        """通过配置创建连接对象"""
        return cls(settings)

    async def exchange_declare(self, exchange: str = None) -> spec.Exchange.DeclareOk:
        return await self._do_query(func_name='exchange_declare',
                                    exchange=exchange,
                                    exchange_type='topic',
                                    durable=True)

    async def queue_declare(self, queue: str = '') -> spec.Queue.DeclareOk:
        return await self._do_query(func_name='queue_declare',
                                    queue=queue,
                                    durable=True,
                                    arguments=self.queue_arg)

    async def get_queue_message_count(self, queue: str):
        """
        获取队列的msg数量
        :param queue:
        :return:
        """
        result = await self.queue_declare(queue)
        return result.message_count

    async def send_msg(self, msg: str, exchange: str, routing_key: str,
                       priority: int) -> typing.Optional[ConfirmationFrameType]:
        """
        发送消息 -> 默认持久化
        :param msg:
        :param exchange: 交换机名称
        :param routing_key: 路由关键字
        :param priority: 优先级 1-5 超过5将失效
        :return:
        """
        properties = spec.Basic.Properties(delivery_mode=2, priority=priority)
        return await self._do_query(
            func_name='basic_publish',
            exchange=exchange,
            routing_key=routing_key,
            body=msg.encode(),
            properties=properties)

    async def recv_msg(self, queue: str = '') -> tuple:
        msg = await self._do_query(func_name='basic_get', queue=queue, no_ack=True)
        if msg:
            method, properties, body, _ = msg
            if method is not None and isinstance(body, bytes):
                return method.delivery_tag, body.decode()
        return None, None

    async def _do_query(self, func_name, **kwargs):
        async with self.__pool.acquire() as channel:
            func = getattr(channel, func_name)
            return await func(**kwargs)

    def session(self) -> ChannelSession:
        """返回一个session 来支持手动ack"""
        return ChannelSession(self.__pool)


async def test():
    handler = AioRabbitMQHandler({})
    while True:
        ack, msg = await handler.recv_msg('candidate_Test')
        print(ack, msg)
        p = await handler.send_msg('hello', 'candidate', 'Test', 2)
        print(p)
        c = await handler.ack(ack)
        print(c)
        time.sleep(1)
        print('周期完成')


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(test())



import json

from bspider.utils.tools import make_sign
from .base_http import BaseHttp


class Request(BaseHttp):

    def __init__(self,
                 url,
                 method='GET',
                 callback='parser',
                 headers=None,
                 params=None,
                 data=None,
                 cookies=None,
                 meta=None,
                 priority: int = 0,
                 proxy=None,
                 allow_redirect: bool = False,
                 timeout: int = 10,
                 verify_ssl: bool = False,
                 errback: str = 'errback',
                 sign: str = None,
                 ignore_filter: bool = False,
                 life_cycle: int = None):
        """
        :param url: 需要请求的链接
        :param method: 请求的方法
        :param callback: Extractor的回调方法
        :param headers: 请求头
        :param params: get请求参数
        :param data: HTTP请求的body参数
        :param cookies: 需要携带的cookie
        :param meta: 元数据，不跟随request发送到服务端
        :param priority: request的优先级
        :param proxy: {'proxy': http://xxx, 'source': xxxxxx}
        :param allow_redirect: 是否重定向
        :param timeout: 超时时间
        :param verify_ssl: 是否校验证书
        :param errback: 解析异常时的回调函数
        :param sign: request唯一标识，用于下载状态回溯
        :param life_cycle: 该请求的生命周期
        """
        self.url = self._set_url(url)
        self.headers = self._set_headers(headers)
        self.cookies = self._set_cookies(cookies)
        self.method = self._set_method(method)
        self.data = self._set_data(data)
        self.params = self._set_params(params)
        self.meta = self._set_meta(meta)
        self.priority = priority
        self.proxy = proxy  # 下载是否需要使用代理
        self.allow_redirect = allow_redirect  # 下载是否需要重定向
        self.timeout = timeout
        self.verify_ssl = verify_ssl
        self.callback = self._set_callback(callback)
        self.errback = self._set_errback(errback)
        self.sign = make_sign(self.url, salt='' if data is None else json.dumps(data)) if sign is None else sign
        self.ignore_filter = ignore_filter # 过滤器识别标识
        self.life_cycle = life_cycle

    @classmethod
    def loads(cls, param: dict):
        return cls(**param)

    def dumps(self):
        return self.__dict__

    def __str__(self):
        return "<Request:%s %s %s>" % (self.method, self.url, self.sign)

    __repr__ = __str__



"""
BSpider - a high-level distributed crawling framework written for Python
"""

__all__ = ['__version__', 'version_info']

# Scrapy version
import pkgutil
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
version_info = tuple(int(v) if v.isdigit() else v
                     for v in __version__.split('.'))
del pkgutil

# Check minimum required Python version
import sys
if sys.version_info < (3, 7):
    print("BSpider %s requires Python 3.7" % __version__)
    sys.exit(1)



# 运行模式，本地开发请使用development，服务部署请使用produce
# development log 输出至终端
# produce log 输出至文件
GLOBAL_PATTERN = 'development'
LOGGER_LEVEL = 'DEBUG'

#################
# MASTER模块配置 #
#################
# 各节点的 AGENT 进程将会通过访问 http://ip:port 和主进程进行通信
# web studio访问地址：http://ip:port
# ****************************************************
# web studio 默认用户: 账号admin，密码admin 请登录后尽快修改
# ****************************************************

# 时区配置 str
TIMEZONE = 'Asia/Shanghai'

# JWT Token 生效时间
WEB_TOKEN_LIVE_TIME = 3600 * 5

# 定时任务进程启动多少工作线程用以执行定时任务
CRON_JOB_THREAD_NUM = 20
# 多久同步一次定时任务信息，单位s, 数字越小效果越好，数字越大性能消耗越小
CRON_JOB_REFRESH_TIME = 3

#################
# AGENT  模块配置 #
#################
# master节点将通过访问 http://ip:port 和 当前节点agent进行通信

#################
# RABBITMQ 配置  #
#################

#################
# supervisor配置 #
#################
SUPERVISOR_RPC_PORT = 9010
SUPERVISOR_RPC = {
    'username': 'bspider_node',
    'password': 'bspider_node',
    'port': SUPERVISOR_RPC_PORT
}

#################
# 消息通知模块配置 #
#################

# *** 内置配置
# 存储抓取任务的表
PROJECT_TABLE = 'bspider_project'
# 存储用户信息的表
USER_TABLE = 'bspider_user'
# 存储鉴权信息的表
AUTH_RULE_TABLE = 'bspider_casbin_auth'
# 节点列表
NODE_TABLE = 'bspider_node'
# worker列表
WORKER_TABLE = 'bspider_worker'
# 节点状态列表
NODE_STATUS_TABLE = 'bspider_node_status'
# 定时任务调度表
CRON_JOB_STORE_TABLE = 'bspider_cron'
# 远程代码仓库表
CODE_STORE_TABLE = 'bspider_customcode'
# 关联表
P2C_TABLE = 'bspider_project_customcode'
# 任务下载状态存储表
DOWNLOADER_STATUS_TABLE = 'bspider_downloader_status'
# 任务解析状态存储表
PARSER_STATUS_TABLE = 'bspider_parser_status'
# rabbitmq - exchange
EXCHANGE_NAME = ('candidate', 'download', 'parse')
# order key
OREDR_KEY = 'order'
# queue arg
QUEUE_ARG = {"x-max-priority": 5, "x-queue-mode": 'lazy'}



import six
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers.base import STATE_STOPPED
from apscheduler.util import undefined
from bspider.bcron.job import MySQLJob


class MySQLScheduler(BackgroundScheduler):

    def add_job(self, func, trigger=None, args=None, kwargs=None, id=None, name=None,
                misfire_grace_time=undefined, coalesce=undefined, max_instances=undefined,
                next_run_time=undefined, jobstore='default', executor='default', status=0,
                description='', replace_existing=False, cron_type='operation', **trigger_args):
        """
        新增了两个属性：
        :param status: 任务状态
        :param description: 任务描述
        :return: str callback msg
        """
        job_kwargs = {
            'trigger': self._create_trigger(trigger, trigger_args),
            'executor': executor,
            'func': func,
            'args': tuple(args) if args is not None else (),
            'kwargs': dict(kwargs) if kwargs is not None else {},
            'id': id,
            'name': name,
            'misfire_grace_time': misfire_grace_time,
            'coalesce': coalesce,
            'max_instances': max_instances,
            'next_run_time': next_run_time,
            # 新增两个属性
            'status': status,
            'description': description,
            'cron_type': cron_type
        }
        job_kwargs = dict((key, value) for key, value in six.iteritems(job_kwargs) if
                          value is not undefined)
        job = MySQLJob(self, **job_kwargs)

        # Don't really add jobs to job stores before the scheduler is up and running
        with self._jobstores_lock:
            if self.state == STATE_STOPPED:
                self._pending_jobs.append((job, jobstore, replace_existing))
                self._logger.info('Adding job tentatively -- it will be properly scheduled when '
                                  'the scheduler starts')
            else:
                self._real_add_job(job, jobstore, replace_existing)

        return job

    def scheduled_job(self, trigger, args=None, kwargs=None, id=None, name=None,
                      misfire_grace_time=undefined, coalesce=undefined, max_instances=undefined,
                      next_run_time=undefined, jobstore='default', executor='default', status=0,
                      description='', **trigger_args):
        """
        新增了两个属性：
        :param status: 任务状态
        :param description: 任务描述
        :return: func
        """
        def inner(func):
            self.add_job(func, trigger, args, kwargs, id, name, misfire_grace_time, coalesce,
                         max_instances, next_run_time, jobstore, executor, status, description,
                         True, **trigger_args)
            return func
        return inner



from flask import Blueprint

from bspider.core.api import auth
from bspider.master.service.code import CodeService
from .validators.code_forms import AddForm, UpdateForm
from .validators import PageForm

code = Blueprint('code_bp', __name__)

code_service = CodeService()


@code.route('/code/<int:code_id>', methods=['GET'])
@auth.login_required
def get(code_id):
    return code_service.get_code(code_id)


@code.route('/code', methods=['GET'])
@auth.login_required
def gets():
    form = PageForm()
    return code_service.get_codes(**form.to_dict())


@code.route('/code', methods=['POST'])
@auth.login_required
def add():
    form = AddForm()
    return code_service.add(**form.to_dict())


@code.route('/code/<int:code_id>', methods=['DELETE'])
@auth.login_required
def delete(code_id):
    return code_service.delete(code_id)


@code.route('/code/<int:code_id>', methods=['PATCH'])
@auth.login_required
def update(code_id):
    form = UpdateForm()
    return code_service.update(code_id, **form.to_dict())



import yaml
from apscheduler.util import obj_to_ref
from pymysql import IntegrityError

from bspider.bcron import do
from bspider.core import Project
from bspider.master import log
from bspider.master.controller.validators.project_form import schema
from bspider.master.service.impl.project_impl import ProjectImpl
from bspider.core.api import BaseService, Conflict, NotFound, PostSuccess, DeleteSuccess, GetSuccess, \
    PatchSuccess, ParameterException, AgentMixIn
from bspider.utils.tools import class_name2module_name, get_crontab_next_run_time


class ProjectService(BaseService, AgentMixIn):

    def __init__(self):
        self.impl = ProjectImpl()

    def add(self, editor: str, config: list, status: int):
        project = Project(config[0],
                          middleware_serializer_method=self.get_module_id_by_name,
                          pipeline_serializer_method=self.get_module_id_by_name)

        log.debug(
            f'pc_obj->pipeline:{project.parser_settings.pipeline}, middleware:{project.downloader_settings.middleware}')

        try:
            with self.impl.mysql_client.session() as session:
                r_config = project.dumps()
                data = {
                    'name': project.project_name,
                    'status': status,
                    'type': 'spider',
                    'group': project.group,
                    'description': project.description,
                    'editor': editor,
                    'rate': project.rate,
                    'config': config[1],
                    'r_config': r_config
                }

                project_id = session.insert(*self.impl.add_project(data), lastrowid=True)
                self.impl.bind_queue(project_id=project_id)
                log.debug(f'bind new project=>{project.project_name} queue success!')
                cids = [[cid for cid in items.keys()][0] for items in r_config['downloader']['middleware']]
                cids.extend([[cid for cid in items.keys()][0] for items in r_config['parser']['pipeline']])
                log.debug(f'code num: {cids}')
                session.insert(*self.impl.add_project_binds(cids, project_id))

                timestamp, next_run_time = get_crontab_next_run_time(project.scheduler_settings.trigger, self.tz)
                value = {
                    'project_id': project_id,
                    'code_id': self.get_module_id_by_name(project.parser_settings.extractor),
                    'type': 'crawl',
                    'trigger': project.scheduler_settings.trigger,
                    'trigger_type': project.scheduler_settings.trigger_type,
                    'func': obj_to_ref(do),
                    'executor': 'thread_pool',
                    'description': project.scheduler_settings.description,
                    'next_run_time': timestamp,
                }
                job_id = session.insert(*self.impl.add_cron_job(value))
                log.info(f'add cron job:{project.project_name} success')

                info = {
                    'project_id': project_id,
                    'name': project.project_name,
                    'rate': project.rate,
                    'config': r_config,
                    'status': status
                }
                node_list = self.impl.get_nodes()
                sign, result = self.op_add_project(node_list, info)
                if not sign:
                    log.error(f'not all node add project:{project.project_name} =>{result}')
                    return Conflict(msg=f'not all node add project:{project.project_name}', data=result, errno=30007)
                log.info(f'add project:{project.project_name} success')
                return PostSuccess(msg=f'add project:{project.project_name} success',
                                   data={'project_id': project_id, 'job_id': job_id})
        except IntegrityError as e:
            if e.args[0] == 1062:
                log.error(f'all project add failed:{project.project_name} =>project is already exist')
                return Conflict(errno=30002, msg='project is already exist')
            raise e

    def update(self, project_id, changes):
        remote_param = dict()
        cron_param = dict()
        log.debug(f'change:{changes}')
        if 'status' in changes:
            remote_param['status'] = changes['status']

        if 'config' in changes:

            infos = self.impl.get_project(project_id)
            if not len(infos):
                return NotFound(msg='project not exist', errno=30001)

            new_project = Project(changes['config'][0],
                                  middleware_serializer_method=self.get_module_id_by_name,
                                  pipeline_serializer_method=self.get_module_id_by_name)
            try:
                old_project = Project(schema(yaml.safe_load(infos[0]['config'])))
            except Exception as e:
                log.warning(f'update error: old project yaml load failed:{e}')
                return Conflict(msg=f'update error: old project yaml load failed:{e}', errno=30003)

            if old_project.project_name != new_project.project_name:
                remote_param['name'] = new_project.project_name
                changes['name'] = new_project.project_name

            if old_project.rate != new_project.rate:
                remote_param['rate'] = new_project.rate
                changes['rate'] = new_project.rate

            if old_project.group != new_project.group:
                changes['group'] = new_project.group

            if old_project.description != new_project.description:
                changes['description'] = new_project.description

            if old_project.scheduler_settings != new_project.scheduler_settings:
                cron_param = new_project.scheduler_settings.dumps()
                cron_param.pop('rate')

            sign = [
                old_project.downloader_settings == new_project.downloader_settings,
                old_project.parser_settings == new_project.parser_settings
            ]
            if False in sign:
                remote_param['config'] = new_project.dumps()
                changes['r_config'] = remote_param['config']

            changes['config'] = changes['config'][1]

        if len(remote_param):
            with self.impl.mysql_client.session() as session:
                session.update(*self.impl.update_project(project_id, changes))
                r_config = changes.get('r_config')
                log.debug(r_config)
                if r_config:
                    cids = [[cid for cid in items.keys()][0] for items in r_config['downloader']['middleware']]
                    cids.extend([[cid for cid in items.keys()][0] for items in r_config['parser']['pipeline']])
                    log.debug(f'code num: {cids}')
                    # 20191224 bug修复，修复无法删除module引用
                    session.delete(*self.impl.delete_project_binds(project_id))
                    session.insert(*self.impl.add_project_binds(cids, project_id))
                    if len(cron_param):
                        session.insert(*self.impl.update_cron_job_by_project_id(project_id, cron_param))
                sign, result = self.op_update_project(self.impl.get_nodes(), project_id, remote_param)
                if not sign:
                    log.warning(f'not all node update project:project_id->{project_id} =>{result}')
                    return Conflict(msg=f'not all node update this project change', data=result, errno=30007)
        else:
            with self.impl.mysql_client.session() as session:
                session.update(*self.impl.update_project(project_id, changes))
                if len(cron_param):
                    session.insert(*self.impl.update_cron_job_by_project_id(project_id, cron_param))

        log.info(f'update project:project_id->{project_id} success')
        return PatchSuccess(msg=f'update project success')

    def delete(self, project_id):
        with self.impl.mysql_client.session() as session:
            session.delete(*self.impl.delete_project(project_id))
            session.delete(*self.impl.delete_project_binds(project_id))
            session.delete(*self.impl.delete_cron_job(project_id))
            sign, result = self.op_delete_project(self.impl.get_nodes(), project_id)
            if not sign:
                log.error(f'all project delete failed:{project_id} =>{result}')
                raise Conflict(msg='project delete failed', data=result, errno=30007)
            log.info(f'delete project:{project_id} success')
            if self.impl.unbind_queue(project_id):
                return DeleteSuccess()
            else:
                raise Conflict(msg='project\'s queue is not exist', errno=30001)

    def get(self, project_id):
        infos = self.impl.get_project(project_id)
        for info in infos:
            self.datetime_to_str(info)

        if len(infos):
            return GetSuccess(data=infos[0])
        else:
            return NotFound(msg='project not exist', errno=30001)

    def gets(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_projects(page, limit, search, sort)

        for info in infos:
            self.datetime_to_str(info)

        return GetSuccess(
            msg='get user list success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })

    def get_module_id_by_name(self, cls_name: str) -> int:
        module_type = class_name2module_name(cls_name).split('_')[-1]
        data = self.impl.get_module_id_by_name_and_type(cls_name, module_type)
        if len(data):
            return data[0]['id']
        raise Conflict(
            msg=f'lack of necessary middleware <{cls_name}>!',
            errno=30003)



from wtforms import StringField, IntegerField

from bspider.core.api import BaseForm, ParamRequired


class RegisterForm(BaseForm):
    worker_id = IntegerField(validators=[ParamRequired()])
    coroutine_num = IntegerField(validators=[ParamRequired()])
    worker_type = StringField(validators=[ParamRequired()])



from bspider.agent import log
from bspider.core import AgentCache
from bspider.core.api import BaseService, GetSuccess, PostSuccess, DeleteSuccess, PatchSuccess


class CodeService(BaseService):

    def __init__(self):
        self.cache = AgentCache()

    def add_code(self, code_id, content):
        self.cache.set_code(code_id, content)
        log.info(f'add code:code_id->{code_id} success')
        return PostSuccess()

    def get_codes(self):
        return GetSuccess(data=self.cache.get_codes())

    def get_code(self, code_id):
        return GetSuccess(data=self.cache.get_code(code_id))

    def update_code(self, code_id, changes):
        project = changes.pop('project')
        self.cache.update_code(code_id, changes, project)
        log.info(f'update code:code_id->{code_id} success')
        return PatchSuccess()

    def delete_code(self, code_id):
        self.cache.delete_code(code_id)
        log.info(f'delete code:code_id->{code_id} success')
        return DeleteSuccess()



import os
from collections import namedtuple

import casbin
from flask import request
from itsdangerous import TimedJSONWebSignatureSerializer, BadSignature, SignatureExpired

import bspider
from bspider.config import FrameSettings
from bspider.core.api.exception import AuthFailed, Forbidden
from bspider.utils.logger import LoggerPool

__log = LoggerPool().get_logger(key='api_auth', fn='auth', module='api_auth')
__key = FrameSettings()['WEB_SECRET_KEY']

User = namedtuple('User', ['user_id', 'role', 'token'])
e = casbin.Enforcer(os.path.join(bspider.__path__[0], 'config', 'rbac_model.conf'),
                    os.path.join(bspider.__path__[0], 'config', 'rbac_policy.csv'))


def make_token(user_id, role):
    s = TimedJSONWebSignatureSerializer(__key, expires_in=7200)
    return s.dumps({'user_id': user_id, 'role': role}).decode('ascii')


def verify_token(token):
    s = TimedJSONWebSignatureSerializer(__key)
    try:
        data = s.loads(token)
    except BadSignature:
        raise AuthFailed(msg='token is invalid', errno=10002)
    except SignatureExpired:
        raise AuthFailed(msg='token is expired', errno=10003)
    if e.enforce(data['role'], request.method.upper(), request.blueprint):
        __log.info(
            'user:{}-{} request:{} {} pass verify'.format(data['user_id'], data['role'], request.endpoint, request.method))
        return User(data['user_id'], data['role'], token)
    __log.info(
        'user:{}-{} request:{} {} auth failed'.format(data['user_id'], data['role'], request.endpoint, request.method))
    raise Forbidden(msg='Unauthorized operation', errno=10004)



"""
整个command模块都参考了scrapy 的command模块
https://github.com/scrapy/scrapy/blob/master/scrapy/cmdline.py
https://github.com/scrapy/scrapy/tree/master/scrapy/commands
"""
import inspect
import optparse
import sys
import os

from bspider.commands import BSpiderCommand
from bspider.config import FrameSettings
from bspider.utils.conf import init_platform_env, PLATFORM_NAME_ENV, BSPIDER_VERSION_ENV
from bspider.utils.exceptions import UsageError
from bspider.utils.importer import walk_modules


class CommandLine(object):

    def __init__(self):
        self.cmds = self.__get_commands_from_module()
        print(self.cmds)
        self.args = list()

    @staticmethod
    def __get_commands_from_module(module='bspider.commands'):
        cmds = dict()
        for mod in walk_modules(module):
            for obj in vars(mod).values():
                if inspect.isclass(obj) and \
                        issubclass(obj, BSpiderCommand) and \
                        obj.__module__ == mod.__name__ and \
                        not obj == BSpiderCommand:
                    cmds[obj.__module__.split('.')[-1]] = obj()
        return cmds

    @staticmethod
    def __print_header():
        if os.environ.get(PLATFORM_NAME_ENV):
            print(f"BSpider {os.environ[BSPIDER_VERSION_ENV]} - platform {os.environ[PLATFORM_NAME_ENV]}")
        else:
            print(f"BSpider {os.environ[BSPIDER_VERSION_ENV]} - no active platform")

    @property
    def command_name(self):
        i = 0
        for arg in self.args[1:]:
            if not arg.startswith('-'):
                del self.args[i]
                return arg
            i += 1

    @init_platform_env
    def execute(self, argv):
        self.args = argv
        command_name = self.command_name
        if not command_name:
            self.__print_header()
            print("Usage: bspider <command> [options] [args]")
            print("Available commands:")
            for cmdname, cmdclass in sorted(self.cmds.items()):
                print("  %-13s %s" % (cmdname, cmdclass.short_desc()))
            print('Use "bspider <command> -h" to see more info about a command')
            sys.exit(0)

        if not command_name in self.cmds:
            self.__print_header()
            print("Unknown command: %s\n" % command_name)
            sys.exit(2)

        if command_name == 'mkplatform' or command_name == 'version':
            frame_settings = dict()
        elif os.environ.get(PLATFORM_NAME_ENV):
            frame_settings = FrameSettings()
        else:
            self.__print_header()
            print('Please use bspider mkplatform <platform_name> to create a platform')
            sys.exit(2)

        command = self.cmds[command_name]
        parser = optparse.OptionParser(formatter=optparse.IndentedHelpFormatter(), conflict_handler='resolve')
        parser.prog = command_name
        parser.usage = f"bspider {command_name} {command.syntax()}"
        parser.description = command.long_desc()
        command.add_options(parser)
        command.frame_settings = frame_settings
        try:
            command.run(*parser.parse_args(self.args))
        except UsageError:
            parser.print_help()
            sys.exit(2)
        sys.exit(command.exitcode)


def execute(argv=None):
    """
    执行终端命令
    :param cmdname:
    :param args:
    :return:
    """
    CommandLine().execute(sys.argv if argv is None else argv)


if __name__ == '__main__':
    print(os.listdir('.'))
    print(execute(['cmdline.py', 'startspider', 'c1_baidu_map']))



from bspider.core.api import BaseService, RabbitMQMixIn, GetSuccess, DeleteSuccess


class RabbitMQService(BaseService, RabbitMQMixIn):

    def get_project_queue_info(self, project_id: int):
        queue_infos = self.op_get_project_queue_detail(project_id)
        result = dict()
        for queue_info in queue_infos:
            result[queue_info['name'].replace(f'_{project_id}', '')] = {
                'total': [queue_info['messages'], queue_info['messages_details']['rate']],
                'unacknowledged': [queue_info['messages_unacknowledged'], queue_info['messages_unacknowledged_details']['rate']],
                'ready': [queue_info['messages_ready'], queue_info['messages_ready_details']['rate']],
            }
        return GetSuccess(msg='get queue info success', data=result)

    def purge_project_queue(self, project_id: int):
        self.op_purge_queue_msg(project_id)
        return DeleteSuccess()





from bspider.core.api import BaseImpl
from bspider.master import log


class CodeImpl(BaseImpl):

    def add_code(self, data):
        fields, values = self.make_fv(data)
        sql = f"insert into {self.code_table} set {fields};"
        return sql, values, True

    def get_nodes(self):
        sql = f'select `ip` from {self.node_table} where `status` = 1;'
        return [info['ip'] for info in self.mysql_client.select(sql)]

    def get_project_by_code_id(self, cid):
        sql = f'SELECT `p2c`.`project_id` AS `id`, `project`.`name` AS `name` FROM {self.project_table} AS `project` ' \
              f'LEFT JOIN {self.p2c_table} AS `p2c` ON `project`.`id`=`p2c`.`project_id` ' \
              f'WHERE `p2c`.`customcode_id`={cid};'
        return self.mysql_client.select(sql)

    def update_code(self, code_id, data):
        fields, values = self.make_fv(data)
        sql = f"update {self.code_table} set {fields} where `id` = {code_id};"
        log.debug(f'SQL:{sql}, {values}')
        return sql, values

    def delete_code(self, code_id):
        sql = f'delete from {self.code_table} where `id`={code_id};'
        return sql, ()

    def show_bind_project_code(self, project_id):
        sql = f'SELECT `code`.`name` AS `code_name`,`code`.`type` AS `code_type` ' \
              f'FROM {self.table_name} AS `code` ' \
              f'LEFT JOIN bspider_project_customcode AS `p2c` ON `code`.`id` = `p2c`.`customcode_id` ' \
              f'WHERE `p2c`.`project_id` = {project_id};'
        return self.mysql_client.select(sql)

    def get_code(self, code_id):
        sql = f'SELECT `project`.`name` AS `project_name`,`project`.`id` AS `project_id`,' \
              f'`code`.`name` AS `name`,`code`.`id` AS `id`,`code`.`type`,' \
              f'`code`.`description`,`code`.`content`,`code`.`editor` ' \
              f'FROM {self.code_table} AS `code` ' \
              f'LEFT JOIN bspider_project_customcode AS `p2c` ON `code`.`id`=`p2c`.`customcode_id` ' \
              f'LEFT JOIN bspider_project AS `project` ON `project`.`id`=`p2c`.`project_id` ' \
              f'WHERE `code`.`id`={code_id}'
        return self.mysql_client.select(sql)

    def get_code_by_type(self, code_type):
        sql = f'select `id`, `name`, `description`, `type`, `editor`, `content` ' \
              f'from {self.code_table} ' \
              f'where `type` = "{code_type}";'
        return self.mysql_client.select(sql)

    def get_codes(self, page, limit, search, sort):
        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `name`, `description`, `type`, `editor`, `create_time`, `update_time` ' \
                  f'from {self.code_table} where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `name`, `description`, `type`, `editor`, `create_time`, `update_time` ' \
                  f'from {self.code_table} order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.code_table)



from .request import Request
from .response import Response

ERROR_REQUEST = Request('bspider://error')

ERROR_RESPONSE = Response(
        url='bspider://error',
        status=599,
        request=ERROR_REQUEST
    )

__all__ = ['Request', 'Response', 'ERROR_RESPONSE']



"""
所有和 Agent 进行交互的操作收缩到 node 蓝图下
1. 提供注册接口用于注册新 node
2. 删除接口删除节点 node 停止节点下所有work => 删除worker 表中节点对应的所有信息 => 停止Agent进程(supervisor)
4. 查询接口查询 节点信息 返回节点和节点下已经注册的所有worker信息
5. 停止节点
6. 启动节点

5. 在节点下启动并在master注册一个worker 启动进程 => 成功后注册进表单
6. 在节点下停止一个 worker 停止进程 status 设置为0
7. 在节点下重启一个 worker 停止进程、启动进程
8. 删除一个 worker 停止进程并删除注册信息
9. 查询 worker 得到节点下所有 worker 信息
"""
from flask import Blueprint

from bspider.core.api import auth
from .validators import PageForm
from .validators.node_forms import AddNodeForm, UpdateNodeForm, AddWorkerForm, UpdateWorkerForm
from bspider.master.service.node import Node

node = Blueprint('node_bp', __name__)

node_service = Node()


@node.route('/node', methods=['POST'])
@auth.login_required
def add_node():
    form = AddNodeForm()
    return node_service.add_node(**form.to_dict())


@node.route('/node/<int:node_id>', methods=['DELETE'])
@auth.login_required
def delete_node(node_id):
    return node_service.delete_node(node_id)


@node.route('/node/<int:node_id>', methods=['PATCH'])
@auth.login_required
def update_node(node_id):
    form = UpdateNodeForm()
    return node_service.update_node(node_id, **form.to_dict())


@node.route('/node', methods=['GET'])
@auth.login_required
def get_nodes():
    form = PageForm()
    return node_service.get_nodes(**form.to_dict())


@node.route('/node/<int:node_id>', methods=['GET'])
@auth.login_required
def get_node(node_id):
    return node_service.get_node(node_id)


@node.route('/worker', methods=['POST'])
@auth.login_required
def add_worker():
    form = AddWorkerForm()
    return node_service.add_worker(form.ip.data, form.name.data, form.type.data, form.description.data, int(form.status.data))


@node.route('/worker/<int:worker_id>', methods=['DELETE'])
@auth.login_required
def delete_worker(worker_id):
    return node_service.delete_worker(worker_id)


@node.route('/worker/<int:worker_id>', methods=['PATCH'])
@auth.login_required
def update_worker(worker_id):
    form = UpdateWorkerForm()
    return node_service.update_worker(worker_id, **form.to_dict())


@node.route('/worker', methods=['GET'])
@auth.login_required
def get_workers():
    """获取全部worker信息"""
    form = PageForm()
    return node_service.get_workers(**form.to_dict())


@node.route('/worker/<int:worker_id>', methods=['GET'])
@auth.login_required
def get_worker(worker_id):
    return node_service.get_worker(worker_id)



"""
整个 command 模块都参考了scrapy 的 commands模块
这里引用了scrapy 中的代码：https://github.com/scrapy/scrapy/blob/master/scrapy/commands/__init__.py
"""
import os
import re
from optparse import OptionGroup
from os.path import join
from shutil import ignore_patterns, copy2, copystat

import bspider

IGNORE = ignore_patterns('*.pyc', '.svn', '__pycache__')


class BSpiderCommand(object):
    # default settings to be used for this command instead of global defaults
    frame_settings = {}

    exitcode = 0

    def _is_valid_name(self, name):
        def _spider_exists(module_name):
            dirs = [d for d in os.listdir('.') if os.path.isdir(os.path.join('.', d))]
            return module_name in dirs

        if not re.search(r'^[_a-zA-Z]\w*$', name):
            print('Error: names must begin with a letter and contain'
                  ' only\nletters, numbers and underscores')
        elif _spider_exists(name):
            print('Error: Module %r already exists' % name)
        else:
            return True
        return False

    def _copytree(self, src, dst):
        """
        copy platform template
        """
        ignore = IGNORE
        names = os.listdir(src)
        ignored_names = ignore(src, names)

        if not os.path.exists(dst):
            os.makedirs(dst)

        for name in names:
            if name in ignored_names:
                continue

            srcname = os.path.join(src, name)
            dstname = os.path.join(dst, name)
            if os.path.isdir(srcname):
                self._copytree(srcname, dstname)
            else:
                copy2(srcname, dstname)
        copystat(src, dst)

    def __init__(self):
        self.settings = {}  # set in scrapy.cmdline

    def syntax(self):
        """
        Command syntax (preferably one-line). Do not include command name.
        """
        return ""

    def short_desc(self):
        """
        A short description of the command
        """
        return ""

    def long_desc(self):
        """A long description of the command. Return short description when not
        available. It cannot contain newlines, since contents will be formatted
        by optparser which removes newlines and wraps text.
        """
        return self.short_desc()

    def help(self):
        """An extensive help for the command. It will be shown when using the
        "help" command. It can contain newlines, since not post-formatting will
        be applied to its contents.
        """
        return self.long_desc()

    def add_options(self, parser):
        """
        Populate option parse with options available for this command
        """
        group = OptionGroup(parser, "Global Options")
        parser.add_option_group(group)

    def process_options(self, args, opts):
        pass

    def run(self, args, opts):
        """
        Entry point for running commands
        """
        raise NotImplementedError

    @property
    def templates_dir(self):
        try:
            return self.settings['TEMPLATES_DIR']
        except KeyError:
            return join(bspider.__path__[0], 'templates')



"""
下载器基类
执行中间件方法
"""
import asyncio
import sys
import traceback

import aiohttp
from aiohttp import ClientResponse

from bspider.core import Project
from bspider.http import Request, Response
from bspider.utils.exceptions import DownloaderError
from bspider.utils.importer import import_module_by_code
from bspider.utils.logger import LoggerPool
from bspider.utils.sign import Sign


class AsyncDownloader(object):

    def __init__(self, project: Project, sign: Sign, log_fn: str):
        """传入下载器的配置文件"""
        self.sign = sign
        self.project_name = project.project_name
        self.project_id = project.project_id

        self.log = LoggerPool().get_logger(
            key=f'project_downloader->{self.project_id}', fn=log_fn, module='downloader', project=self.project_name)
        # 加载重试次数
        self.retry_times = project.downloader_settings.max_retry_times
        self.log.debug(f'{self.project_name} init retry time from settings: {self.retry_times}')

        self.ignore_retry_http_code = set(project.downloader_settings.ignore_retry_http_code)
        self.ignore_retry_http_code.add(599)
        self.ignore_retry_http_code.add(200)
        self.log.debug(
            f'{self.project_name} init accept response code from settings: {self.ignore_retry_http_code}')

        self.mws = []
        for middleware in project.downloader_settings.middleware:
            for cls, params in middleware.items():
                cls_name, code = cls
                if isinstance(code, str):
                    mod = import_module_by_code(cls_name, code)
                else:
                    mod = code
                if mod and hasattr(mod, cls_name):
                    try:
                        self.mws.append(
                            getattr(mod, cls_name)(project, {**project.global_settings, **params}, self.log))
                        self.log.info(f'success load: <{self.project_name}:{cls_name}>!')
                    except Exception as e:
                        raise DownloaderError('<%s:%s> middleware init failed: %s' % (self.project_name, cls_name, e))

                else:
                    msg = f'<{self.project_name}:{cls_name}> middleware init failed: middleware is invalid!'
                    raise DownloaderError(msg)

    async def download(self, request: Request) -> (Response, bool, str):
        self.log.info(f'===>> start download sign:{request.sign} {request}')
        response = Response(url=request.url, status=599, request=request)
        # 异常占位符
        e_msg = None
        for retry_index in range(self.retry_times):
            is_req = True
            # 下载前中间件
            for mw in self.mws:
                self.log.debug(f'{mw.__class__.__name__} executing process_request')
                result = await mw._exec('process_request', request)

                if isinstance(result, Request):
                    request = result
                    continue
                elif result is None:
                    continue
                elif isinstance(result, Response):
                    is_req = False
                    response = result
                    break

            try:
                if is_req:
                    response = await self.__do(request)
            except Exception as e:
                tp, msg, tb = sys.exc_info()
                e_msg = ''.join(traceback.format_exception(tp, msg, tb))
                e.with_traceback(tb)
                self.log.exception(e_msg)
                # 执行下载异常中间件
                for mw in self.mws:
                    self.log.debug(f'{mw.__class__.__name__} executing process_response')
                    result = await mw._exec('process_exception', request, e, response)
                    if result is None:
                        self.log.info(f'===>> complete download sign:{request.sign} {request}')
                        return response, True, e_msg
                continue

            for mw in self.mws:
                self.log.debug(f'{mw.__class__.__name__} executing process_response')
                result = await mw._exec('process_response', request, response)
                if result is None:
                    break

            if response.status in self.ignore_retry_http_code:
                self.log.debug(f'response.status == {response.status}: url->{request.url} ignore retry')
                self.log.info(f'===>> complete download sign:{request.sign} {request}')
                return response, True, None
            self.log.info(f'Retry download: url->{request.url} status->{response.status} time:{retry_index + 1}')
        self.log.info(f'===>> complete download sign:{request.sign} {request}')
        return response, True, e_msg

    async def __assemble_response(self, response: ClientResponse, request: Request) -> Response:
        # 这里只处理 str 类型的数据
        text = await response.text(errors='ignore')
        return Response(
            url=str(response.url),
            status=response.status,
            headers=dict(response.headers),
            request=request,
            cookies={i.key: i.value for i in response.cookies.values()},
            text=text
        )

    async def __do(self, req: Request) -> Response:
        # ClientSession 在每次请求都要关闭，所以使用上下文管理器进行管理
        temp_timeout = aiohttp.ClientTimeout(total=20, connect=req.timeout)
        self.log.debug(f'set time out {temp_timeout.connect}')
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method=req.method,
                    url=req.url,
                    headers=req.headers,
                    params=req.params,
                    # post 参数，get时为 None
                    data=req.data,
                    cookies=req.cookies,
                    # 是否允许重定向
                    allow_redirects=req.allow_redirect,
                    timeout=temp_timeout,
                    proxy=None if not isinstance(req.proxy, dict) else req.proxy.get('proxy'),
                    # ssl验证
                    ssl=req.verify_ssl,
            ) as resp:
                # 挂起等待下载结果
                self.log.debug(f'aiohttp finish {temp_timeout.connect}')
                return await self.__assemble_response(resp, req)


if __name__ == '__main__':
    bd = AsyncDownloader()

    req = Request(
        url='https://www.baidu.com/',
        method='GET',
        timeout=10
    )

    loop = asyncio.get_event_loop()
    tasks = [asyncio.ensure_future(bd.__do(req))]

    loop.run_until_complete(asyncio.wait(tasks))
    loop.close()



from wtforms import StringField, IntegerField, FloatField
from wtforms.validators import length

from bspider.core.api import BaseForm, ParamRequired


class AddNodeForm(BaseForm):
    ip = StringField(validators=[ParamRequired(), length(min=7, max=15)])
    name = StringField(validators=[ParamRequired(), length(min=3, max=30)])
    description = StringField(validators=[ParamRequired()])
    cpu_num = IntegerField(validators=[ParamRequired()])
    mem_size = FloatField(validators=[ParamRequired()])
    disk_size = FloatField(validators=[ParamRequired()])
    port = IntegerField(validators=[ParamRequired()])

class UpdateNodeForm(BaseForm):
    name = StringField()
    status = IntegerField()
    description = StringField()
    cpu_num = IntegerField()
    mem_size = IntegerField()
    disk_size = IntegerField()


class AddWorkerForm(BaseForm):
    ip = StringField(validators=[ParamRequired(), length(min=7, max=15)])
    name = StringField(validators=[ParamRequired(), length(min=3, max=30)])
    type = StringField(validators=[ParamRequired()])
    description = StringField(validators=[ParamRequired()])
    status = IntegerField(default=1)


class UpdateWorkerForm(BaseForm):
    ip = StringField()
    name = StringField()
    type = StringField()
    description = StringField()
    status = IntegerField()



import asyncio
import inspect
import os
from importlib import import_module
from os.path import abspath
from queue import Queue

import yaml

from bspider.config import FrameSettings
from bspider.core import Project
from bspider.master.controller.validators.project_form import schema
from bspider.downloader.async_downloader import AsyncDownloader
from bspider.parser.async_parser import AsyncParser
from bspider.http import Request
from bspider.utils.conf import PLATFORM_NAME_ENV
from bspider.utils.exceptions import ModuleExistError, ParserError
from bspider.utils.database import MysqlClient
from bspider.utils.logger import LoggerPool
from bspider.utils.sign import Sign
from bspider.utils.tools import class_name2module_name


class Debuger(object):
    """继承此类进行BSpider 框架的开发调试"""
    frame_settings = FrameSettings()

    # 最大下载次数 下载超过此数量个 Request 调试进程会终止
    max_follow_url_num = 10

    # proority fix rabbitmq 数字越大优先级越高，使用Python优先队列模拟的时候需要进行优先级翻转

    def __init__(self):
        self.debug_log_fn = 'terminal'
        self.log = LoggerPool().get_logger(
            key=self.project_name,
            module='debuger',
            project=self.project_name,
            fn=self.debug_log_fn
        )
        self.mysql_client = MysqlClient.from_settings(self.frame_settings.get('WEB_STUDIO_DB'))

        self.max_priority = self.frame_settings['QUEUE_ARG'].get('x-max-priority', 5)
        self.priority_queue = [Queue() for _ in range(self.max_priority)]

        with open(abspath('settings-development.yml'), encoding='utf8') as f:
            self.project = Project(schema(yaml.safe_load(f)),
                                   middleware_serializer_method=self.load_module,
                                   pipeline_serializer_method=self.load_module)
            self.project.project_id = 0

        self.parser = AsyncParser(self.project, Sign(), self.debug_log_fn)
        self.log.info('init parser success')
        self.downloader = AsyncDownloader(self.project, Sign(), self.debug_log_fn)
        self.log.info('init downloader success')
        self.__cur_download_num = 0
        self.log.info('send Request to queue')
        for extractor in self.parser.pipes:
            count = 0
            if not extractor.__class__.__name__.endswith('Extractor'):
                continue
            for request in extractor.start_url():
                self.put(request)
                self.log.info(f'project:project_id->{self.project.project_id} success send a request->{request.sign}')
                count += 1
                if count > 100:
                    self.log.warning(f'Debug url over 100 ... Ignore remaining URLs')
                    break
            break

    @property
    def project_name(self):
        return os.path.basename(os.getcwd())

    def put(self, request: Request):
        """向优先队列中存入req"""
        try:
            self.priority_queue[request.priority].put(request)
            self.log.debug(f'success send request {request}')
        except IndexError:
            ParserError(
                f'At req:%s, requests.priority=%s must between 1~%s' % (request.url, request.priority, self.max_priority))

    def get(self):
        """从优先队列中取出req，若空则返回None"""
        for i in range(self.max_priority - 1, -1, -1):
            if self.priority_queue[i].empty():
                continue
            return self.priority_queue[i].get()

    def start(self):
        """调试入口"""
        tasks = [asyncio.ensure_future(self.__debug())]
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.wait(tasks))

        self.log.info('debuger finished')

    async def __debug(self):
        while True:
            request = self.get()
            if request is None:
                self.log.debug('candidate queue is empty')
                break

            self.log.info(f'start download url:{request.url}')
            response, sign, _ = await self.downloader.download(request)
            self.__cur_download_num += 1
            if sign:
                self.log.info(f'start parser url:{response.url} status:{response.status}')
                reqs = await self.parser.parse(response)
                for req in reqs:
                    self.put(req)

            if self.__cur_download_num > self.max_follow_url_num:
                self.log.info(f'debuger follow url {self.max_follow_url_num} second')
                break

    def load_module(self, cls_name):
        module_name = class_name2module_name(cls_name)
        module_type = module_name.split('_')[-1]
        if module_type == 'extractor':
            module_path = f'{os.environ[PLATFORM_NAME_ENV]}.projects.{self.project.project_name}.extractor'
        else:
            module_path = f'{os.environ[PLATFORM_NAME_ENV]}.{module_type}.{module_name}'

        try:
            mod = import_module(module_path)

            for obj in vars(mod).values():
                if inspect.isclass(obj):
                    if obj.__name__ == cls_name:
                        self.log.debug(f'success find {module_type}:{obj.__name__} from local')
                        return cls_name, mod
        except ModuleNotFoundError:
            pass

        sql = "select `content` from `{}` where `name`=%s;".format(self.frame_settings['CODE_STORE_TABLE'])
        info = self.mysql_client.select(sql, (cls_name))
        if len(info):
            self.log.debug(f'success find module:{cls_name} from remote')
            return cls_name, info[0]['content']
        else:
            raise ModuleExistError('load module from remote failed %s is not exists' % (cls_name))



'''
1. 实现一个validator 校验yaml配置文件的正确性
2. 返回ProjectSettings对象
    - DownloaderSettings
        - MiddleWareSettings:
    - PipelineSettings
        - PipelineSettings
    - BCronSettings
    - SchedulerSettings
'''
import operator
import types

from bspider.utils.exceptions import ProjectSettingsError


class Project(object):

    def __init__(self,
                 settings: dict,
                 project_id_serializer_method = None,
                 middleware_serializer_method= None,
                 pipeline_serializer_method = None):
        self.__project_name = settings['project_name']
        self.__project_id = None
        self.__global_settings = settings['global_settings']
        self.__group = settings['group']
        self.__description = settings['description']
        self.__downloader = settings['downloader']
        self.__parser = settings['parser']
        self.__scheduler = settings['scheduler']

        # 初始化序列化器 序列化器完成 id<->name的转化
        self.__project_id_serializer_method = project_id_serializer_method
        self.__pipeline_serializer_method = pipeline_serializer_method
        self.__middleware_serializer_method = middleware_serializer_method

    @property
    def project_name(self):
        return self.__project_name

    @property
    def project_id(self):
        if self.__project_id is not None:
            return self.__project_id
        if self.__project_id_serializer_method:
            self.__project_id = self.__project_id_serializer_method(self.project_name)
            return self.__project_id
        raise ProjectSettingsError('project_id_serializer_method is None')

    @project_id.setter
    def project_id(self, project_id):
        self.__project_id = project_id

    @property
    def rate(self):
        return self.scheduler_settings.rate

    @property
    def group(self):
        return self.__group

    @property
    def description(self):
        return self.__description

    @property
    def global_settings(self):
        return self.__global_settings

    @property
    def downloader_settings(self):
        return DownloaderSettings(
            max_retry_times=self.__downloader['max_retry_times'],
            ignore_retry_http_code=self.__downloader['ignore_retry_http_code'],
            middleware=self.__downloader['middleware'],
            serializer_method=self.__middleware_serializer_method
        )

    @property
    def parser_settings(self):
        return ParserSettings(
            pipeline=self.__parser['pipeline'],
            serializer_method=self.__pipeline_serializer_method
        )

    @property
    def scheduler_settings(self):
        return SchedulerSettings(settings=self.__scheduler)

    def dumps(self):
        '''dumps 方法需要serializer 方法支持'''
        return {
            'project_name': self.project_name,
            'group': self.group,
            'description': self.description,
            'global_settings': self.__global_settings,
            'downloader': self.downloader_settings.dumps(),
            'parser': self.parser_settings.dumps(),
            'scheduler': self.scheduler_settings.dumps()
        }


class DownloaderSettings(object):

    def __init__(self,
                 max_retry_times: int,
                 ignore_retry_http_code: list,
                 middleware: list,
                 serializer_method: types.MethodType = None):
        self.__max_retry_times = max_retry_times
        self.__ignore_retry_http_code = ignore_retry_http_code
        self.__middleware = middleware
        self.__serializer_method = serializer_method

    @property
    def ignore_retry_http_code(self):
        return self.__ignore_retry_http_code

    @property
    def max_retry_times(self):
        return self.__max_retry_times

    @property
    def middleware(self):
        return MiddlewareSettings(self.__middleware, self.__serializer_method, ['middleware'])

    def dumps(self):
        return {
            'max_retry_times': self.__max_retry_times,
            'ignore_retry_http_code': self.__ignore_retry_http_code,
            'middleware': list(self.middleware)
        }


class ParserSettings(object):

    def __init__(self,
                 pipeline: list,
                 serializer_method: types.MethodType = None):
        self.__pipeline = pipeline
        self.__serializer_method = serializer_method



    @property
    def extractor(self):
        for pipeline in self.__pipeline:
            for key, _ in pipeline.items():
                if key.endswith('Extractor'):
                    return key


    @property
    def pipeline(self):
        return PipelineSettings(self.__pipeline, self.__serializer_method, ['pipeline', 'extractor'])

    def dumps(self):
        return {
            'pipeline': list(self.pipeline)
        }


class SchedulerSettings(object):

    def __init__(self, settings: dict):
        self.__rate = settings['rate']
        self.__trigger_type = settings['trigger_type']
        self.__trigger = settings['trigger']
        self.__description = settings['description']

    @property
    def rate(self):
        return self.__rate

    @property
    def trigger(self):
        return self.__trigger

    @property
    def trigger_type(self):
        return self.__trigger_type

    @property
    def description(self):
        return self.__description

    def dumps(self):
        return {
            'rate': self.__rate,
            'trigger': self.__trigger,
            'description': self.__description
        }

    def __eq__(self, other):
        return other.trigger == self.__trigger \
               and other.trigger_type == self.__trigger_type \
               and other.description == self.description \
               and other.rate == self.rate


class BaseModuleSettings(object):

    def __init__(self, data: list, serializer_method: types.MethodType, module_type: list):
        self.__data = data
        self.__serializer_method = serializer_method
        self.__end = len(data)
        self.__cur = -1
        self.module_type = module_type

    def __iter__(self):
        return self

    def __next__(self) -> (str, dict):
        self.__cur += 1
        if self.__cur == self.__end:
            raise StopIteration()

        for cls, params in self.__data[self.__cur].items():
            if self.__serializer_method:
                return {self.__serializer_method(cls): params}
            return {cls: params}
            # raise ProjectSettingsError('serializer_method is None')

    def __eq__(self, other):
        return operator.eq(self.__data, other)

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        result = list()
        for module in self.__data:
            for cls, _ in module.items():
                result.append(str(cls))
        return '<{}: {} items>'.format(self.__class__.__name__, '->'.join(result))

    __str__ = __repr__


class MiddlewareSettings(BaseModuleSettings):
    pass


class PipelineSettings(BaseModuleSettings):
    pass



from .base_impl import BaseImpl
from .base_service import BaseService
from .base_form import BaseForm, ParamRequired
from .auth import auth
from .exception import *
from .api_response import *
from .remote_mixin import AgentMixIn, MasterMixIn, RabbitMQMixIn


from bspider.core.api import BaseImpl
from bspider.master import log


class CronImpl(BaseImpl):

    def add_job(self, data: dict) -> int:
        data['status'] = 1
        fields, values = self.make_fv(data)
        sql = f"insert into {self.cron_table} set {fields}"
        log.debug(f'SQL:{sql}')
        return self.mysql_client.insert(sql, values, lastrowid=True)

    def update_job(self, job_id: int, data: dict) -> int:
        data['status'] = 2
        fields, values = self.make_fv(data)
        sql = f"update {self.cron_table} set {fields} where `id` = '{job_id}';"
        log.debug(f'SQL:{sql}')
        return self.mysql_client.update(sql, values)

    def get_job(self, job_id: int) -> list:
        sql = f'select `id`, `project_id`, `code_id`, `type`, `trigger`, `trigger_type`, `next_run_time`, ' \
              f'`description`, `create_time`, `update_time` ' \
              f'from `{self.cron_table}` where `id`={job_id};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def get_jobs(self, page: int, limit: int, search: str, sort: str) -> (list, int):

        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `project_id`, `code_id`, `type`, `trigger`, `trigger_type`, `next_run_time`, ' \
                  f'`description`, `create_time`, `update_time` ' \
                  f'from `{self.cron_table}` ' \
                  f'where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `project_id`, `code_id`, `type`, `trigger`, `trigger_type`, `next_run_time`, ' \
                  f'`description`, `create_time`, `update_time` ' \
                  f'from `{self.cron_table}` ' \
                  f'order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.cron_table)

    def delete_job(self, job_id: int) -> int:
        sql = f"update {self.cron_table} set `status`=%s where `id` = '{job_id}';"
        log.debug(f'SQL:{sql}')
        return self.mysql_client.update(sql, (3))



import json
import logging

from bspider.utils.system import System


def get_json_formatter(**kwargs):
    """返回指定的log formatter"""
    log_from_work = {
        'time': '%(asctime)s',
        'pid': '%(process)d',
        'filename': '%(filename)s:%(lineno)d',
        'level': '%(levelname)s',
        'node': System.ip_msg,
        'msg': '%(message)s',
    }
    for key, value in kwargs.items():
        if key in log_from_work:
            continue
        log_from_work[key] = value
    return logging.Formatter(json.dumps(log_from_work))


def get_stream_formatter(**kwargs):
    arg = ' '.join([f'{value}' for key, value in kwargs.items() if key == 'module'])
    log_from_work = f'[%(asctime)s {arg} %(process)d %(filename)s:%(lineno)s] %(levelname)s: %(message)s'
    return logging.Formatter(log_from_work)

def get_file_formatter(**kwargs):
    arg = ' '.join([f'{key}={value}' for key, value in kwargs.items()])
    log_from_work = f'[%(asctime)s %(process)d %(filename)s:%(lineno)s] %(levelname)s: ({arg}) %(message)s'
    return logging.Formatter(log_from_work)



from wtforms import IntegerField, StringField

from bspider.core.api import BaseForm, ParamRequired


class AddForm(BaseForm):
    project_id = IntegerField(validators=[ParamRequired()])
    name = StringField(validators=[ParamRequired()])
    config = StringField(validators=[ParamRequired()])
    rate = IntegerField(validators=[ParamRequired()])
    status = IntegerField(validators=[ParamRequired()])


class UpdateForm(BaseForm):
    name = StringField()
    config = StringField()
    rate = IntegerField()
    status = IntegerField()



from bspider.core.api import BaseImpl
from bspider.master import log


class ToolsImpl(BaseImpl):

    def get_node_list(self):
        sql = f"select `id`, `name`, `ip` from `{self.node_table}`;"
        return self.mysql_client.select(sql)

    def get_code_list(self, code_type):
        if code_type:
            sql = f"select `id`, `name`, `editor` from `{self.code_table}` where `type`='{code_type}';"
        else:
            sql = f"select `id`, `name`, `editor` from `{self.code_table}`;"
        return self.mysql_client.select(sql)


    def get_downloader_exception(self, project_id):
        """目前是自动获取8小时内，前4条错误信息"""
        sql = f'SELECT `sign`,`method`, `url`, `status`, `url_sign`,' \
              f'`exception`,`create_time` as crawl_time ' \
              f'FROM {self.downloader_status_table} ' \
              f'WHERE `create_time` > now()-INTERVAL 8 HOUR ' \
              f'AND `project_id`="{project_id}" ' \
              f'AND `status` NOT BETWEEN 900 AND 999 ' \
              f'AND `status` !=200 ' \
              f'ORDER BY `create_time` DESC LIMIT 4'
        return self.mysql_client.select(sql)

    def get_parser_exception(self, project_id):
        sql = f'SELECT `sign`,`method`,`url`,`status`,`url_sign`,' \
              f'`exception`,`create_time` as crawl_time ' \
              f'FROM {self.parser_status_table} ' \
              f'WHERE `create_time` > now()-INTERVAL 8 HOUR ' \
              f'AND `project_id`="{project_id}" ' \
              f'AND `exception` is not null ' \
              f'GROUP BY `exception` ORDER BY `create_time` DESC LIMIT 4'
        return self.mysql_client.select(sql)

    def get_crawl_detail(self, value, source):
        if source == 'parser':
            table = self.parser_status_table
        else:
            table = self.downloader_status_table

        sql = f'select `project_id`, `project_name`,`sign`,`method`,`data` ,`url`,`status`,`url_sign`,`response`,' \
              f'`exception`,`create_time` as crawl_time ' \
              f'FROM {table} ' \
              f'WHERE  `sign`="{value}"'

        return self.mysql_client.select(sql)

    def get_sign_by_url(self, url):
        sql = f'select `sign` from {self.downloader_status_table} ' \
              f'where `url_sign`=md5("{url}") ' \
              f'ORDER BY `create_time` DESC LIMIT 1'
        info = self.mysql_client.select(sql)
        if len(info):
            log.info('url:{} -> sign:{}'.format(url, info[0]['sign']))
            return info[0]['sign']
        log.info(f'can\'t find url:{url} sign')

    def get_node_detail(self):
        sql = f'SELECT count(1) AS `node`,' \
              f'SUM(`mem_size`) AS `memory`, ' \
              f'SUM(`disk_size`) AS `disk`, ' \
              f'convert(SUM(`cpu_num`), signed) AS `cpu` ' \
              f'from {self.node_table} where `status`=1'
        return self.mysql_client.select(sql)

    def get_exception_project(self):
        sql = f'select `id`, `name`, `editor`, `rate`, `status` from {self.project_table} where `status` = -1'
        return self.mysql_client.select(sql)

    def get_user_by_id(self, id):
        """暂时先这样，后面增加复杂密码验证"""
        sql = f'select `id`, `identity`, `username`, `role`, `email`, `phone`, `status`, `create_time`, `update_time`  ' \
              f'from {self.user_table} where `id`=%s;'
        return self.mysql_client.select(sql, id)




import hashlib

import time

import pymysql
from DBUtils.PooledDB import PooledDB
from pymysql.cursors import DictCursor

from bspider.utils import singleton
from bspider.utils.database.mysql.session import DBSession


@singleton
class MysqlPoolFactory(object):
    MYSQL_DEFAULT_CHARSET = 'utf8'

    def __init__(self):
        self.__pools = dict()

    def __del__(self):
        for key, pool in self.__pools.items():
            pool.close()

    def prepare_params(self, *args, **kwargs):
        mysql_params = {
            'host': None,
            'port': None,
            'user': None,
            'password': None,
            'db': None,
            'charset': self.MYSQL_DEFAULT_CHARSET,
        }
        if args or kwargs:
            for k, v in dict(*args, **kwargs).items():
                if k in mysql_params:
                    mysql_params[k] = v
        return mysql_params

    @staticmethod
    def __make_hash(mysql_config):
        hashstr = ','.join([f'{k}:{v}' for k, v in mysql_config.items()])
        hashcode = hashlib.md5(hashstr.encode('utf8')).hexdigest()
        return hashcode

    def get_pool(self, creator=pymysql, mincached=0, maxcached=20,
                 maxshared=20, maxconnections=20,
                 *args, **kwargs):
        mysql_params = self.prepare_params(*args, **kwargs)
        hashcode = self.__make_hash(mysql_params)

        if hashcode in self.__pools:
            return self.__pools[hashcode]
        else:
            pool = PooledDB(creator=creator,
                            mincached=mincached, maxcached=maxcached,
                            maxshared=maxshared, maxconnections=maxconnections,
                            *args, **kwargs)
            self.__pools[hashcode] = pool
            return pool


class MysqlClient(object):
    """封装了常用操作的mysql句柄，重试默认3次, 线程安全"""

    def __init__(self, mysql_config):
        self.__pool = MysqlPoolFactory().get_pool(
            host=mysql_config['host'],
            port=mysql_config['port'],
            user=mysql_config['user'],
            password=mysql_config['password'],
            db=mysql_config['db'],
            charset=mysql_config['charset'],
            cursorclass=DictCursor
        )

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    @staticmethod
    def _do_query(func, sql: str, values: tuple = None, retry_times: int = 0,
                  retry_interval: int = 1, lastrowid: bool = False):
        loop_times = retry_times if retry_times > 0 else 3
        while loop_times > 0:
            try:
                if lastrowid:
                    return func(sql, values, lastrowid=lastrowid)
                else:
                    return func(sql, values)
            except pymysql.Error as e:
                errid, errmsg = e.args
                if errid != 2014:
                    raise e
            time.sleep(retry_interval)
        raise pymysql.MySQLError('mysql connection error')

    def select(self, sql: str, values: tuple = None,
               retry_times: int = 0, retry_interval: int = 1) -> list:
        """
        提供重试的查询方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 查询结果
        """
        return self._do_query(self._select, sql,
                              values=values,
                              retry_times=retry_times,
                              retry_interval=retry_interval)

    def insert(self, sql: str, values: tuple = None, retry_times: int = 0,
               retry_interval: int = 1, lastrowid: bool = False) -> int:
        """
        提供重试的插入方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :param lastrowid: 是否返回自增id
        :return: 插入结果
        """
        return self._do_query(self._insert, sql,
                              values=values,
                              retry_times=retry_times,
                              retry_interval=retry_interval,
                              lastrowid=lastrowid)

    def update(self, sql: str, values: tuple = None,
               retry_times: int = 0, retry_interval: int = 1) -> int:
        """
        提供重试的更新方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 更新结果
        """
        return self._do_query(self._update, sql, values=values,
                              retry_times=retry_times,
                              retry_interval=retry_interval)

    def delete(self, sql: str, values: tuple = None,
               retry_times: int = 0, retry_interval: int = 1) -> int:
        """
        提供重试的删除方法方法
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: 删除行数
        """
        return self._do_query(self._delete, sql, values=values,
                              retry_times=retry_times,
                              retry_interval=retry_interval)

    def query(self, sql: str, values: tuple = None,
              retry_times: int = 0, retry_interval: int = 1):
        """
        query
        :param sql: sql语句
        :param values: sql语句中的值
        :param retry_times: 重试次数
        :param retry_interval: 重试间隔
        :return: T
        """
        return self._do_query(self._query, sql, values=values,
                              retry_times=retry_times,
                              retry_interval=retry_interval)

    def session(self) -> DBSession:
        """返回一个session 来支持事务"""
        return DBSession(self.__pool, self.select)

    def _select(self, sql: str, values: tuple = None) -> list:
        conn = None
        cursor = None
        try:
            conn = self.__pool.dedicated_connection()
            cursor = conn.cursor()
            if values is None:
                result_count = cursor.execute(sql)
            else:
                result_count = cursor.execute(sql, values)
            if result_count > 0:
                return cursor.fetchall()
            else:
                return []
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.close()

    def _query(self, sql: str, values: tuple = None, lastrowid: bool = False):
        conn = None
        cursor = None
        try:
            conn = self.__pool.dedicated_connection()
            cursor = conn.cursor()
            if values is None:
                result = cursor.execute(sql)
            else:
                result = cursor.execute(sql, values)

            if lastrowid:
                return cursor.lastrowid
            else:
                return result
        finally:
            if cursor:
                cursor.close()
            if conn:
                conn.commit()
                conn.close()

    def _insert(self, sql: str, values: tuple = None, lastrowid: bool=False) -> int:
        return self._query(sql, values, lastrowid=lastrowid)

    def _update(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)

    def _delete(self, sql: str, values: tuple = None) -> int:
        return self._query(sql, values)



import os
import sys

from lxml import etree

import bspider
from bspider.commands import BSpiderCommand
from bspider.utils.conf import BSPIDER_VERSION_ENV


class Command(BSpiderCommand):

    def syntax(self):
        return ""

    def short_desc(self):
        return "Print BSpider version"

    def add_options(self, parser):
        parser.add_option("--verbose", "-v", dest="verbose", action="store_true",
                          help="also display aiohttp/python/lxml info (useful for bug reports)")

    def run(self, args, opts):
        if not opts.verbose:
            print(f"BSpider {os.environ[BSPIDER_VERSION_ENV]}")
        else:
            lxml_version = ".".join(map(str, etree.LXML_VERSION))
            libxml2_version = ".".join(map(str, etree.LIBXML_VERSION))

            versions = [
                ("bspider", bspider.__version__),
                ("lxml", lxml_version),
                ("libxml2", libxml2_version),
                ("Python", sys.version.replace("\n", "- ")),
            ]
            width = max(len(n) for (n, _) in versions)
            patt = "%-{}s : %s".format(width)
            for name, version in versions:
                print(patt % (name, version))



"""
这里着重参考了scrapy 的response 对象
Response 对象要求实现domps 和 loads 方法来完成 obj <-> 消息之间的转换
"""
import copy
import json

from lxml import etree

from bspider.http import Request
from bspider.http.base_http import BaseHttp


class Response(BaseHttp):
    def __init__(self,
                 url,
                 status,
                 request,
                 headers=None,
                 cookies=None,
                 text=None):
        self.url = self._set_url(url)
        self.headers = self._set_headers(headers)
        self.status = status
        self.text = self._set_text(text)
        self.cookies = self._set_cookies(cookies)
        self.request = self.__set_request(request)
        self.callback = self.request.callback
        self.errback = self.request.errback
        self.sign = self.request.sign
        self.meta = self.request.meta
        self.method = self.request.method

    def __set_request(self, request):
        if isinstance(request, dict):
            return Request.loads(request)
        elif isinstance(request, Request):
            return request
        else:
            raise TypeError("%s request must be dict or bspider.Request object. " % (type(self).__name__))

    def dumps(self):
        """解决序列化的嵌套问题"""
        resp = dict(
            url=self.url,
            status=self.status,
            request=self.request if not isinstance(self.request, Request) else self.request.dumps(),
            headers=self.headers,
            cookies=self.cookies,
            text=self.text
        )
        return resp

    @classmethod
    def loads(cls, param: dict):

        return cls(**param)

    @property
    def json(self):
        return json.loads(self.text)

    @property
    def xpath_selector(self):
        """:return a xpath selector obj"""
        return etree.HTML(self.text)

    def __str__(self):
        return "<Response:%s %s %s>" % (self.method, self.status, self.url)

    __repr__ = __str__



import asyncio
import time

import aiohttp
import requests
from aiohttp import ClientResponse

from bspider.http import Request, Response

proxy = '27.152.204.239:22148'

def get_proxy():
    url = 'http://proxy.baishanglin.top/proxy/'
    proxy = requests.get(url).json()['data']
    print(proxy)
    proxy['proxy'] = 'https://{proxy}'.format(**proxy)
    return proxy

class Test(object):

    async def __assemble_response(self, response: ClientResponse, request: Request) -> Response:
        # 这里只处理 str 类型的数据
        text = await response.text(errors='ignore')
        return Response(
            url=str(response.url),
            status=response.status,
            headers=dict(response.headers),
            request=request,
            cookies={i.key: i.value for i in response.cookies.values()},
            text=text
        )

    async def do(self, req: Request) -> Response:
        """
        执行下载操作
        url': '', # 请求的url 必须
        method': '', # 请求的方法 GET, POST, PUT ,PATCH, OPTIONS, HEAD, DELETE
        request_body': {}, # POST 请求的消息体 字典结构
        cookies': {}, # 请求时携带的cookie 字典结构
        meta': {}, # 请求时携带的上下文信息，通常与本次请求无关
        proxy': str
        allow_redirect': bool, # 下载是否需要重定向
        timeout: int /s
        task_info': {
            name': '', # 任务信息
            task_sign': '', # 一个任务中一个链接中每一次请求的唯一标识
        }
        :param param: dict 下载参数
        :return:
        """
        # sc 在每次请求都要关闭，所以使用上下文管理器进行管理
        temp_timeout = aiohttp.ClientTimeout(connect=req.timeout)
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method=req.method,
                    url=req.url,
                    headers=req.headers,
                    # post 参数，get时为 None
                    data=req.data,
                    cookies=req.cookies,
                    # 是否允许重定向
                    allow_redirects=req.allow_redirect,
                    timeout=temp_timeout,
                    proxy=None if not isinstance(req.proxy, dict) else req.proxy.get('proxy'),
                    # ssl验证
                    ssl=req.verify_ssl,
            ) as resp:
                # 挂起等待下载结果
                return await self.__assemble_response(resp, req)


if __name__ == '__main__':
    bd = Test()

    req = Request(
        url='https://www.baidu.com/',
        method='GET',
        timeout=10,
        proxy= get_proxy()
    )

    loop = asyncio.get_event_loop()
    tasks = [asyncio.ensure_future(bd.do(req))]

    loop.run_until_complete(asyncio.wait(tasks))
    loop.close()



import os
from os.path import join, exists

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_PATH_ENV
from bspider.utils.exceptions import UsageError
from bspider.utils.template import render_templatefile
from bspider.utils.tools import class_name2module_name

TEMPLATES_TO_RENDER = (
    ('middleware.py.tmpl',),
)


class Command(BSpiderCommand):

    def syntax(self):
        return "<middleware name like DemoMiddleware>"

    def short_desc(self):
        return "Create new middleware in current platform"

    def run(self, args, opts):
        if len(args) not in (1, 2):
            raise UsageError('args error')

        middleware_name = args[0]
        file_name = f'{class_name2module_name(middleware_name)}.py'
        project_dir = join(os.environ[PLATFORM_PATH_ENV], 'middleware')
        if exists(join(project_dir, file_name)):
            self.exitcode = 1
            print(f'Error: middleware: {middleware_name} already exists in this platform')
            return

        self._copytree(join(self.templates_dir, 'middleware'), project_dir)
        for paths in TEMPLATES_TO_RENDER:
            tplfile = join(project_dir, *paths)
            new_file = join(project_dir, file_name)
            render_templatefile(tplfile, new_file=new_file, middleware_name=middleware_name)

        print(f"New middleware '{middleware_name}' init success")
        print(f"    template directory:\n    '{self.templates_dir}'")


if __name__ == '__main__':
    Command().run(['Hello_project'])



from bspider.core.api import BaseImpl
from bspider.master import log


class ChartImpl(BaseImpl):

    def get_project_downloader_pv(self, project_id: int = None):
        """获取按小时统计的下载数据"""
        if project_id:
            sql = f"SELECT DATE_FORMAT(`create_time`,'%Y-%m-%d %H:00:00') AS `time`," \
                  f"COUNT(IF (`status` !=200,TRUE,NULL)) AS `exception`," \
                  f"COUNT(*) AS total " \
                  f"FROM {self.downloader_status_table} " \
                  f"WHERE `project_id`={project_id} " \
                  f"GROUP BY `time` ORDER BY `time`;"
        else:
            sql = f"SELECT DATE_FORMAT(`create_time`,'%Y-%m-%d %H:00:00') AS `time`," \
                  f"COUNT(IF (`status` !=200,TRUE,NULL)) AS `exception`," \
                  f"COUNT(*) AS total " \
                  f"FROM {self.downloader_status_table} " \
                  f"GROUP BY `time` ORDER BY `time`;"
        return self.mysql_client.select(sql)

    def get_project_parser_pv(self, project_id: int = None):
        if project_id:
            sql = f"SELECT DATE_FORMAT(`create_time`,'%Y-%m-%d %H:00:00') AS `time`," \
                  f"COUNT(IF (`exception` is not null,TRUE,NULL)) AS `exception`," \
                  f"COUNT(*) AS total " \
                  f"FROM {self.parser_status_table} " \
                  f"WHERE `project_id`={project_id} " \
                  f"GROUP BY `time` ORDER BY `time`;"
        else:
            sql = f"SELECT DATE_FORMAT(`create_time`,'%Y-%m-%d %H:00:00') AS `time`," \
                  f"COUNT(IF (`exception` is not null,TRUE,NULL)) AS `exception`," \
                  f"COUNT(*) AS total " \
                  f"FROM {self.parser_status_table} " \
                  f"GROUP BY `time` ORDER BY `time`;"
        return self.mysql_client.select(sql)

    def get_code_type_detail(self):
        sql = f'select count(1) as `value`, `type` as `name` from {self.code_table} GROUP BY `type`'
        return self.mysql_client.select(sql)

    def get_node_pv(self, node_ip: str = None):
        """获取按小时统计的下载数据"""
        sql = f"SELECT DATE_FORMAT(`create_time`,'%Y-%m-%d %H:00:00') AS `time`," \
              f"avg(`memory`) as `memory`," \
              f"avg(`cpu`) as `cpu`, " \
              f"avg(`disk`) as `disk` " \
              f"FROM {self.node_status_table} " \
              f"WHERE `ip`='{node_ip}' " \
              f"GROUP BY `time` ORDER BY `time`;"
        log.debug(sql)
        return self.mysql_client.select(sql)



"""
这个是实际执行的方法，这个方法接收参数，通过参数决定行为
"""
import json
import sys
import traceback

import yaml

from bspider.config import FrameSettings
from bspider.core import Project
from bspider.http import Request
from bspider.utils.database import MysqlClient
from bspider.utils.logger import LoggerPool
from bspider.utils.importer import import_module_by_code
from bspider.utils.notify import ding
from bspider.utils.rabbitMQ import RabbitMQClient

__frame_settings = FrameSettings()
__mysql_client = MysqlClient.from_settings(__frame_settings['WEB_STUDIO_DB'])
__mq_client = RabbitMQClient(__frame_settings['RABBITMQ_CONFIG'])
__log = LoggerPool().get_logger(key='bcorn-todo', fn='bcorn', module='bcorn')


def do(**kwargs):
    # 模式判断
    if kwargs.get('type') == 'operation':
        run_operation_project(**kwargs)
    elif kwargs.get('type', 'crawl') == 'crawl':
        run_spider_project(**kwargs)
    else:
        __log.warning(f'WARNING: UNKNOW CRON JOB PARAMS: {kwargs}')


def run_spider_project(project_id, code_id, **kwargs):
    """
    执行自动加载脚本的定时任务
    先检查定时任务状态，如果为1，就执行否则跳出
    :param self:
    :return:
    """
    __log.info(f'corn job: {project_id, code_id} run as crawl pattern')
    PROJECT_TABLE = __frame_settings['PROJECT_TABLE']
    sql = f'select `id`, `name`, `r_config` as `config`, `status` from {PROJECT_TABLE} where `id`="{project_id}"'

    infos = __mysql_client.select(sql)
    __log.debug(f'run spider_project cron_job select sql:{sql}')

    if not len(infos):
        __log.warning(f'project is not exist project_id:{project_id}, spider project job:{project_id}-{code_id}')
        ding(f'cron job error: \nproject is not exist project_id:{project_id}')

    info = infos[0]
    if info['status'] != 1:
        __log.warning(f'project_id:{project_id} is not run')
        # ding(f'spider project job:{project_name} is not run')

    run_status, run_msg = run_corn_job_code(
        code_id=code_id,
        project_id=project_id,
        do_type='project',
        config=info['config'])
    __log.debug(info['config'])
    if run_status:
        __log.info(run_msg)
    else:
        __log.error(f'{run_msg}\narg:{kwargs}')
        ding(run_msg, 'crawl task')


def run_operation_project(code_id, **kwargs):
    run_status, run_msg = run_corn_job_code(
        code_id=code_id,
        project_id=0,
        do_type='operation')
    if run_status:
        __log.info(run_msg)
    else:
        __log.error(f'{run_msg}\narg:{kwargs}')
        ding(run_msg, 'operation task')


def run_corn_job_code(code_id, project_id, do_type, config: str = None):
    CODE_STORE_TABLE = __frame_settings['CODE_STORE_TABLE']
    sql = f'select `name`, `content` from {CODE_STORE_TABLE} where `id`="{code_id}"'
    tmp = __mysql_client.select(sql)
    if not len(tmp):
        return False, f'{code_id} is not exist in remote code store'
    content = tmp[0]['content']
    class_name = tmp[0]['name']
    mod = import_module_by_code(class_name, content)
    instance_arg = list()
    if hasattr(mod, class_name):
        if do_type == 'operation':
            instance_arg.append(LoggerPool().get_logger(
                key=f'bcorn-{class_name}',
                fn='bcorn',
                module='bcorn',
                project='operation'))
        else:
            try:
                project = Project(yaml.safe_load(config))
                project.project_id = project_id
                instance_arg.append(project)
                instance_arg.append(project.global_settings)
                instance_arg.append(LoggerPool().get_logger(
                    key=project.project_name,
                    fn='bcorn',
                    module='bcorn',
                    project=project.project_name))

                requests = getattr(mod, class_name)(*instance_arg).start_url()
                for request in requests:
                    send_request(request, project)
                return True, f'{project_id}-{code_id} run succeed'
            except Exception:
                tp, msg, tb = sys.exc_info()
                e_msg = '> '.join(traceback.format_exception(tp, msg, tb))
                return False, f'{project_id}-{code_id}:\n > {e_msg}'

        try:
            getattr(mod, class_name)(*instance_arg).execute_task()
            return True, f'{project_id}-{code_id} run succeed'
        except Exception:
            tp, msg, tb = sys.exc_info()
            e_msg = '> '.join(traceback.format_exception(tp, msg, tb))
            return False, f'{project_id}-{code_id} cron job run failed:\n > {e_msg}'

    return False, f'{project_id}-{code_id} don\'t have {class_name}'

def send_request(request: Request, project: Project):
    """发送request 到待下载队列"""
    # 这里dump方法使用了浅拷贝，会影响一部分性能
    data = json.dumps(request.dumps())
    __mq_client.send_msg(
            __frame_settings['EXCHANGE_NAME'][0],
            str(project.project_id),
            data,
            request.priority)
    __log.info(f'project:project_id->{project.project_id} success send a request->{request.sign}')
    __log.debug(f'project:project_id->{project.project_id} Request: {data}')
    return True


import os
import string


def render_templatefile(path, new_file=None, **kwargs):
    """模板文件渲染"""
    with open(path, 'rb') as fp:
        raw = fp.read().decode('utf8')

    content = string.Template(raw).substitute(**kwargs)

    if new_file:
        render_path = new_file
    else:
        render_path = path[:-len('.tmpl')] if path.endswith('.tmpl') else path
    with open(render_path, 'wb') as fp:
        fp.write(content.encode('utf8'))
    if path.endswith('.tmpl'):
        os.remove(path)



from bspider.core.api import BaseImpl
from bspider.master import log
from bspider.utils.database.mysql import prepare_insert_sql


class NodeImpl(BaseImpl):

    def get_all_projects(self):
        sql = f'select `id` as `project_id`, `name`, `status`, `r_config` as `config`, `rate` ' \
              f'from {self.project_table} where `type`= "spider"'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def get_all_codes(self):
        sql = f'select `id` as `code_id`, `content` from {self.code_table} where `type` in ("pipeline", "middleware", "extractor");'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def get_all_workers(self, ip):
        sql = f'select `id` as `worker_id`, `name`, `type`, `coroutine_num` ' \
              f'from {self.worker_table} where `ip` = "{ip}" and `status` = 1;'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def add_node(self, data):
        sql, values = prepare_insert_sql(self.node_table, data, auto_update=True)
        log.debug(f'SQL:{sql} {values}')
        return self.mysql_client.insert(sql, values)

    def delete_node(self, node_id, get_sql=False):
        """a common delete func"""
        sql = f"delete from {self.node_table} where `id`={node_id};"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql,
        return self.mysql_client.delete(sql)

    def delete_worker_by_id(self, worker_id, get_sql=False):
        sql = f"delete from {self.worker_table} where `id` = '{worker_id}';"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql,
        return self.mysql_client.delete(sql)

    def delete_worker_by_ip(self, node_ip, get_sql=False):
        sql = f"delete from {self.worker_table} where `ip` = '{node_ip}';"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql,
        return self.mysql_client.delete(sql)

    def update_node(self, node_id, data, get_sql=False):
        fields, values = BaseImpl.make_fv(data)
        sql = f"update {self.node_table} set {fields} where `id` = {node_id};"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql, values
        return self.mysql_client.update(sql, values)

    def get_node(self, node_id):
        sql = f'select `id`, `ip`, `name`, `description`, `cpu_num`, ' \
              f'`mem_size`, `disk_size`, `status`, `port`, `create_time`, `update_time` ' \
              f'from {self.node_table} where `id`={node_id}'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def get_nodes(self, page, limit, search, sort):
        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `ip`, `name`, `description`, `cpu_num`, ' \
                  f'`mem_size`, `disk_size`, `status`, `port`, `create_time`, `update_time` ' \
                  f'from {self.node_table} where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `ip`, `name`, `description`, `cpu_num`, ' \
                  f'`mem_size`, `disk_size`, `status`, `port`, `create_time`, `update_time` ' \
                  f'from {self.node_table} order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.node_table)

    def add_worker(self, data, get_sql=False):
        """注册一个节点"""
        fields, values = BaseImpl.make_fv(data)
        sql = f"insert into {self.worker_table} set {fields};"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql, values
        return self.mysql_client.insert(sql, values)

    def update_worker(self, worker_id, data, get_sql=False):
        fields, values = BaseImpl.make_fv(data)
        sql = f"update {self.worker_table} set {fields} where `id` = {worker_id};"
        log.debug(f'SQL:{sql}')
        if get_sql:
            return sql, values
        return self.mysql_client.update(sql, values)

    def get_worker(self, worker_id):
        sql = f'select `id`, `ip`, `name`, `type`, `description`, `coroutine_num`, `status`, `create_time`, `update_time` from {self.worker_table} where `id`={worker_id};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql)

    def get_workers(self, page, limit, search, sort):
        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `ip`, `name`, `type`, `description`, `coroutine_num`, `status`, `create_time`, `update_time` ' \
                  f'from {self.worker_table} where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `ip`, `name`, `type`, `description`, `coroutine_num`, `status`, `create_time`, `update_time` ' \
                  f'from {self.worker_table} order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.worker_table)



import datetime
import json

import requests
from urllib3 import disable_warnings

disable_warnings()

from bspider.config import FrameSettings
from bspider.utils.system import System


def ding(msg, title='', at=None):
    """
    发送钉钉报警
    :param msg: str 钉钉报警的信息
    :param at: list at 谁
    :return: None
    """
    cur_time = datetime.datetime.today().strftime('%Y-%m-%d %H:%M:%S')
    data = {
        'msgtype': 'markdown',
        "markdown": {
            "title": f"【BSpider】{title}",
            "text": f"#### 【BSpider】{title}\n{msg}\n> ###### {cur_time} alt {System.ip_msg}"
        },
        "at": {
            "atMobiles": at,
            "isAtAll": False
        }
    }
    headers = {
        'Content-Type': 'application/json'
    }

    return requests.post(
        url=FrameSettings()['DING'],
        headers=headers,
        data=json.dumps(data),
        verify=False
    )
